-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("achievementTracker", "AceEvent-3.0", "AceHook-3.0")

-----------------------
--- Library loading ---
-----------------------
local LATr = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-AchievementTracker")
local LSM = LibStub("LibSharedMedia-3.0")
local QTIP = LibStub("LibQTip-1.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local CreateFont, CreateFrame, SetFont = CreateFont, CreateFrame, SetFont
local type, unpack, format, pairs = type, unpack, string.format, pairs
local tonumber, floor, mod = tonumber, floor, mod
local GetAchievementInfo, GetAchievementLink = GetAchievementInfo, GetAchievementLink
local GetAchievementCategory, GetCategoryInfo, GetCategoryNumAchievements = GetAchievementCategory, GetCategoryInfo, GetCategoryNumAchievements
local GetAchievementNumCriteria, GetAchievementCriteriaInfo = GetAchievementNumCriteria, GetAchievementCriteriaInfo

-----------------
--- Variables ---
-----------------
module.modName = LATr["Achievement Tracker"]
local _
local prdb, chdb, options
local numTrackedAchievement = 0
local category, achievements = {}, {}
local achievementTracker, trackerAnchor, usedTracker
local booleanCriteriaTracked, maxNumberOfBooleanCriteria, maxNumberOfCompletedBooleanCriteria = 0, 0, 0
local anchorPosition = {[1] = "TOPLEFT", [2] = "TOP", [3] = "TOPRIGHT", [4] = "BOTTOMLEFT", [5] = "BOTTOM", [6] = "BOTTOMRIGHT"}
local outlines = {[""] = LATr["None"], ["OUTLINE"] = LATr["Outline"], ["THICKOUTLINE"] = LATr["Thick Outline"]}
local achievementPosition = {[1] = LATr["Above"], [2] = LATr["Below"]}
local cATrFontTQ2 = CreateFont("cATrFontTQ2")
local scATrFontTQ2 = CreateFont("scATrFontTQ2")
local aATrFontTQ2 = CreateFont("aATrFontTQ2")
local acATrFontTQ2 = CreateFont("acATrFontTQ2")
local COLOR_COPPER = "eda55f"
local COLOR_SILVER = "c7c7cf"
local COLOR_GOLD = "ffd700"
local hideInCombat = nil
local interactivityList = {
	["chatLink"] = LATr["Insert achievement link in chat"],
	["achievementTooltip"] = LATr["Display of achievement tooltip"],
	["removeTrackedAchievement"] = LATr["Remove achievement from tracker"],
	["displayAchievement"] = LATr["Diplay achievement frame"],
	["nothing"] = LATr["Do nothing"],
}

----------------------
-- Table recycling ---
----------------------
local tablePool = {}
setmetatable(tablePool, {__mode = "kv"})

local function new()
	local t = next(tablePool) or {}
	tablePool[t] = nil
	return t
end

local function del(t)
	if type(t) == "table" then
		for k, v in pairs(t) do
			if type(v) == "table" then
				v = del(v)
			end
			t[k] = nil
		end
		t[true] = true
		t[true] = nil
		setmetatable(t, nil)
		tablePool[t] = true
	end
	return nil
end

----------------------------
--- Tracker Anchor Frame ---
----------------------------
local anchorSetPoint = {
	[1] = {point = "TOPLEFT", relPoint = "BOTTOMLEFT", x = 0, y = 24},
	[2] = {point = "TOP", relPoint = "BOTTOM", x = 0, y = 24},
	[3] = {point = "TOPRIGHT", relPoint = "BOTTOMRIGHT", x = 0, y = 24},
	[4] = {point = "BOTTOMLEFT", relPoint = "TOPLEFT", x = 0, y = -24},
	[5] = {point = "BOTTOM", relPoint = "TOP", x = 0, y = -24},
	[6] = {point = "BOTTOMRIGHT", relPoint = "TOPRIGHT", x = 0, y = -24}}

local anchor = CreateFrame("Button", nil, UIParent)
	anchor:SetHeight(24)
	anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
	anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
	anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
	local text = anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
	text:SetPoint("CENTER")
	text:SetText(LATr["Achievement Tracker"])
	anchor:SetWidth(text:GetStringWidth() + 8)
	anchor:SetMovable(true)
	anchor:RegisterForDrag("LeftButton")
	anchor:SetClampedToScreen(true)
	anchor:SetScript("OnDragStart", function(self)
		self:StartMoving()
	end)
	anchor:SetScript("OnDragStop", function(self)
		self:StopMovingOrSizing()
		prdb.point, prdb.x, prdb.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
		trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
		module:updateAchievementTracker()
	end)

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		point = "CENTER",
		x = 0,
		y = 0,
		showInQuestTracker = false,
		positionInQuestTracker = 1,
		showAnchor = true,
		anchorPosition = 2,
		toggleTracker = true,
		hideTrackerInCombat = true,
		hideCompletedCriteria = true,
		showByCategory = true,
		categoryFont = LSM.DefaultMedia.font,
		categoryFontSize = 12,
		categoryFontOutline = "",
		categoryCollapseButtonSize = 24,
		subCategoryFont = LSM.DefaultMedia.font,
		subCategoryFontSize = 12,
		subCategoryFontOutline = "",
		subCategoryCollapseButtonSize = 24,
		achievementFont = LSM.DefaultMedia.font,
		achievementFontSize = 10,
		achievementFontOutline = "",
		displayAchievementImage = true,
		achievementImageSize = 24,
		achievementCriteriaFont = LSM.DefaultMedia.font,
		achievementCriteriaFontSize = 8,
		achievementCriteriaFontOutline = "",
		numBooleanCriteriaColumns = 2,
		maxWidth = 300,
		statusMaxWidth = 100,
		hasMaxHeight = true,
		maxHeight = 600,
		subCategoryCellPadding = 5,
		achievementCellPadding = 10,
		criteriaCellPadding = 15,
		rightCellPadding = 5,
		scale = 1,
		backDropColor = {0.09, 0.09, 0.09, 1},
		borderColor = {1, 1, 1, 1},
		disableInteraction = false,
		overTooltip = true,
		mouseRightShift = "chatLink",
		mouseRightControl = "nothing",
		mouseRightAlt = "nothing",
		mouseRight = "achievementTooltip",
		mouseLeftShift = "removeTrackedAchievement",
		mouseLeftControl = "nothing",
		mouseLeftAlt = "nothing",
		mouseLeft = "displayAchievement",
		mouseMiddleShift = "nothing",
		mouseMiddleControl = "nothing",
		mouseMiddleAlt = "nothing",
		mouseMiddle = "nothing",
	},
}

---------------
--- Options ---
---------------
local optGetter, optSetter, questTrackerSetter, colorGetter, colorSetter, catFontSetter, subCatFontSetter, achFontSetter, achCritFontSetter, anchorSetter, combatSetter, interactionDisabled
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		QTIP:Release(usedTracker)
		usedTracker = nil
		achievementTracker = nil
		module:updateAchievementTracker()
	end
	function questTrackerSetter(info, value)
		prdb[info[#info]] = value
		module:updateAchievementTracker()
		if not value then
			core.cbh:Fire("updateQuestsTracker")
		end
	end
	function colorGetter(info)
		return unpack(prdb[info[#info]])
	end
	function colorSetter(info, arg1, arg2, arg3, arg4)
		prdb[info[#info]][1] = arg1
		prdb[info[#info]][2] = arg2
		prdb[info[#info]][3] = arg3
		prdb[info[#info]][4] = arg4
		module:updateAchievementTracker()
	end
	function catFontSetter(info, value)
		prdb[info[#info]] = value
		cATrFontTQ2:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
		module:updateAchievementTracker()
	end
	function subCatFontSetter(info, value)
		prdb[info[#info]] = value
		scATrFontTQ2:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
		module:updateAchievementTracker()
	end
	function achFontSetter(info, value)
		prdb[info[#info]] = value
		aATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achievementFontOutline)
		module:updateAchievementTracker()
	end
	function achCritFontSetter(info, value)
		prdb[info[#info]] = value
		acATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
		module:updateAchievementTracker()
	end
	function anchorSetter(info, value)
		prdb[info[#info]] = value
		if value then
			anchor:Show()
		else
			anchor:Hide()
		end
		module:updateAchievementTracker()
	end
	function combatSetter(info, value)
		prdb[info[#info]] = value
		hideInCombat = nil
		if value then
			core.RegisterCallback(module, "outOfCombat")
			core.RegisterCallback(module, "inCombat")
		else
			core.UnregisterCallback(module, "outOfCombat")
			core.UnregisterCallback(module, "inCombat")
		end
		module:updateAchievementTracker()
	end
	function interactionDisabled()
		if prdb.disableInteraction then
			return true
		else
			return false
		end
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			childGroups = "tab",
			name = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				display = {
					type = "group",
					name = LATr["Display"],
					args = {
						showInQuestTracker = {
							name = LATr["Show achievements in quests tracker"],
							type = "toggle",
							set = questTrackerSetter,
							width = "full",
							order = 1,
						},
						positionInQuestTracker = {
							name = LATr["Position of achievements relative to quests in the tracker"],
							type = "select",
							values = achievementPosition,
							set = questTrackerSetter,
							width = "full",
							order = 2,
						},
						showAnchor = {
							name = LATr["Show achievement tracker anchor"],
							type = "toggle",
							set = anchorSetter,
							width = "full",
							order = 3,
						},
						anchorPosition = {
							name = LATr["Anchor position"],
							type = "select",
							values = anchorPosition,
							width = "full",
							order = 4,
						},
						toggleTracker = {
							name = LATr["Display tracker"],
							type = "toggle",
							width = "full",
							order = 5,
						},
						hideTrackerInCombat = {
							name = LATr["Hide achievement tracker in combat"],
							type = "toggle",
							set = combatSetter,
							width = "full",
							order = 6,
						},
						hideCompletedCriteria = {
							name = LATr["Hide completed achievement criteria"],
							type = "toggle",
							width = "full",
							order = 7,
						},
						showByCategory = {
							name = LATr["Display achievements by category and subcategory"],
							type = "toggle",
							width = "full",
							order = 8,
						},
					},
				},
				fonts = {
					type = "group",
					name = LATr["Tracker fonts"],
					args = {
						category = {
							type = "group",
							name = LATr["Achievement category"],
							set = catFontSetter,
							inline = true,
							order = 1,
							args = {
								categoryFont = {
									name = LATr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								categoryFontSize = {
									name = LATr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								categoryFontOutline = {
									name = LATr["Font outline"],
									type = "select",
									values = outlines,
									width = "full",
									order = 3,
								},
							},
						},
						subCategory = {
							type = "group",
							name = LATr["Achievement sub category"],
							set = subCatFontSetter,
							inline = true,
							order = 2,
							args = {
								subCategoryFont = {
									name = LATr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								subCategoryFontSize = {
									name = LATr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								subCategoryFontOutline = {
									name = LATr["Font outline"],
									type = "select",
									values = outlines,
									width = "full",
									order = 3,
								},
							},
						},
						achievement = {
							type = "group",
							name = LATr["Achievement"],
							set = achFontSetter,
							inline = true,
							order = 3,
							args = {
								achievementFont = {
									name = LATr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								achievementFontSize = {
									name = LATr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								achievementFontOutline = {
									name = LATr["Font outline"],
									type = "select",
									values = outlines,
									width = "full",
									order = 3,
								},
							},
						},
						achievementCriteria = {
							type = "group",
							name = LATr["Achievement criteria"],
							set = achCritFontSetter,
							inline = true,
							order = 4,
							args = {
								achievementCriteriaFont = {
									name = LATr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								achievementCriteriaFontSize = { 
									name = LATr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								achievementCriteriaFontOutline = {
									name = LATr["Font outline"],
									type = "select",
									values = outlines,
									width = "full",
									order = 3,
								},
							},
						},
					},
				},
				appearance = {
					type = "group",
					name = LATr["Tracker appearance"],
					args = {
						scale = {
							name = LATr["Achievement tracker scale"],
							type = "range",
							min = 0,
							max = 1,
							step = 0.1,
							width = "full",
							order = 1,
						},
						backDropColor = {
							name = LATr["Achievement tracker background color"],
							type = "color",
							hasAlpha = true,
							get = colorGetter,
							set = colorSetter,
							width = "full",
							order = 2,
						},
						borderColor = {
							name = LATr["Achievement tracker border color"],
							type = "color",
							hasAlpha = true,
							get = colorGetter,
							set = colorSetter,
							width = "full",
							order = 3,
						},
						hasMaxHeight = {
							name = LATr["Set achievements tracker maximum height after which it will be scrollable"],
							type = "toggle",
							width = "full",
							order = 4,
						},
						maxHeight = {
							name = LATr["Achievement tracker max height"],
							type = "range",
							min = 100,
							max = 2000,
							step = 5,
							width = "full",
							order = 5,
						},
						categoryCollapseButtonSize = {
							name = LATr["Category collapse Button size"],
							type = "range",
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 6,
						},
						subCategoryCollapseButtonSize = {
							name = LATr["Sub category collapse Button size"],
							type = "range",
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 7,
						},
						displayAchievementImage = {
							name = LATr["Display achievements icon"],
							type = "toggle",
							width = "full",
							order = 8,
						},
						achievementImageSize = {
							name = LATr["Achievement image size"],
							type = "range",
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 9,
						},
						subCategoryCellPadding = {
							name = LATr["Sub category cell padding"],
							type = "range",
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 10,
						},
						achievementCellPadding = {
							name = LATr["Achievement cell padding"],
							type = "range",
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 11,
						},
						criteriaCellPadding = {
							name = LATr["Achievement criteria cell padding"],
							type = "range",
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 12,
						},
						rightCellPadding = {
							name = LATr["Right-most cell right padding to avoid cell clipping"],
							type = "range",
							min = 0,
							max = 10,
							step = 1,
							width = "full",
							order = 13,
						},
						maxWidth = {
							name = LATr["Achievement tracker column max width"],
							type = "range",
							min = 50,
							max = 600,
							step = 5,
							width = "full",
							order = 14,
						},
						statusMaxWidth = {
							name = LATr["Achievement Tracker status column max width"],
							type="range",
							min = 30,
							max = 200,
							step = 5,
							width = "full",
							order = 15,
						},
						numBooleanCriteriaColumns = {
							name = LATr["Number of columns in the achievement tracker for boolean criteria"],
							type = "range",
							min = 1,
							max = 8,
							step = 1,
							width = "full",
							order = 16,
						},
					},
				},
				interaction = {
					type = "group",
					name = LATr["Tracker interaction"],
					args = {
						disableInteraction = {
							type = "toggle",
							name = LATr["Completely disable tracker interaction, should not be used when display by category is used or if you wish to make the it scrollable."],
							width = "full",
							order = 1,
						},
						overTooltip = {
							name = LATr["Display achievement tooltip on mouse over"],
							type = "toggle",
							width = "full",
							hidden = interactionDisabled,
							order = 2,
						},
						mouseRightShift = {
							name = LATr["Action on right shift click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 3,
						},
						mouseRightControl = {
							name = LATr["Action on right control click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 4,
						},
						mouseRightAlt = {
							name = LATr["Action on right alt click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 5,
						},
						mouseRight = {
							name = LATr["Action on right click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 6,
						},
						mouseLeftShift = {
							name = LATr["Action on left shift click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 7,
						},
						mouseLeftControl = {
							name = LATr["Action on left control click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 8,
						},
						mouseLeftAlt = {
							name = LATr["Action on left alt click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 9,
						},
						mouseLeft = {
							name = LATr["Action on left click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 10,
						},
						mouseMiddleShift = {
							name = LATr["Action on middle shift click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 11,
						},
						mouseMiddleControl = {
							name = LATr["Action on middle control click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 12,
						},
						mouseMiddleAlt = {
							name = LATr["Action on middle alt click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 13,
						},
						mouseMiddle = {
							name = LATr["Action on middle click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 14,
						},
					},
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("achievementTracker", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	self:SetEnabledState(core:getModuleEnabled("achievementTracker"))
	core:registerModuleOptions("achievementTracker", getOptions, module.modName)
	chdb.achievementTracked = chdb.achievementTracked or {}
	chdb.categoryDisplay = chdb.categoryDisplay or {}
	self:SecureHook("AddTrackedAchievement")
	self:SecureHook("RemoveTrackedAchievement")
	self:SecureHook("GetNumTrackedAchievements")
	self:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
	self:RegisterEvent("ACHIEVEMENT_EARNED")
	LSM.RegisterCallback(self, "LibSharedMedia_Registered", "updateUsedMedia")
end

function module:OnEnable()
	cATrFontTQ2:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
	scATrFontTQ2:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
	aATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achievementFontOutline)
	acATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	core.RegisterCallback(self, "updateAchievementTracker")
	if prdb.hideTrackerInCombat then
		core.RegisterCallback(self, "outOfCombat")
		core.RegisterCallback(self, "inCombat")
	end
	for aid in pairs(chdb.achievementTracked) do
		local _, _, _, completed = GetAchievementInfo(aid)
		if not completed then
			AddTrackedAchievement(aid, true)
		else
			chdb.achievementTracked[aid] = nil
		end
	end
	if AchievementFrame then
		AchievementFrameAchievements_ForceUpdate()
	end
	module:updateAchievementTracker()
end

function module:OnDisable()
	anchor:Hide()
	QTIP:Release(usedTracker)
	usedTracker = nil
	achievementTracker = nil
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
	chdb = self.db.char
	if not self:IsEnabled() then return end
	cATrFontTQ2:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
	scATrFontTQ2:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
	aATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achievementFontOutline)
	acATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if prdb.hideTrackerInCombat then
		core.RegisterCallback(self, "outOfCombat")
		core.RegisterCallback(self, "inCombat")
	end
end

-----------------
--- Callbacks ---
-----------------
function module:inCombat(event)
	hideInCombat = true
	if achievementTracker then achievementTracker:Hide() end
end

function module:outOfCombat(event)
	hideInCombat = nil
	module:updateAchievementTracker()
end

function module:updateUsedMedia(event, medicType, key)
	if medicType == "font" then
		local updateTracker = nil
		if key == prdb.categoryFont then
			updateTracker = true
			cATrFontTQ2:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
		end
		if key == prdb.subCategoryFont then
			updateTracker = true
			scATrFontTQ2:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
		end
		if key == prdb.achievementFont then
			updateTracker = true
			aATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achievementFontOutline)
		end
		if key == prdb.achievementCriteriaFont then
			updateTracker = true
			acATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
		end
		if updateTracker then
			module:updateAchievementTracker()
		end
	end
end

---------------------------
--- Main Module Program ---
---------------------------
local function formatMoney(money)
	local gold = floor(money / 10000)
	local silver = mod(floor(money / 100), 100)
	local copper = mod(floor(money), 100)
	local fMoney
	if gold > 0 then
		if silver > 0 then
			if copper > 0 then
				fMoney = format("|cff%s%d|r.|cff%s%d|r.|cff%s%d|r", COLOR_GOLD, gold, COLOR_SILVER, silver, COLOR_COPPER, copper)
			else
				fMoney = format("|cff%s%d|r.|cff%s%d|r", COLOR_GOLD, gold, COLOR_SILVER, silver)
			end
		elseif copper > 0 then
			fMoney = format("|cff%s%d|r.|cff%s%d|r", COLOR_GOLD, gold, COLOR_COPPER, copper)
		else
			fMoney = format("|cff%s%d|r", COLOR_GOLD, gold)
		end
	elseif silver > 0 then
		if copper > 0 then 
			fMoney = format("|cff%s%d|r.|cff%s%d|r", COLOR_SILVER, silver, COLOR_COPPER, copper)
		else
			fMoney = format("|cff%s%d|r", COLOR_SILVER, silver)
		end
	else
		fMoney = format("|cff%s%d|r", COLOR_COPPER, copper)
	end
	return fMoney
end

--------------
--- Events ---
--------------
function module:ACHIEVEMENT_EARNED(event, aid)
	if achievements[aid] then
		RemoveTrackedAchievement(aid)
	end
end

function module:TRACKED_ACHIEVEMENT_UPDATE(event, aid)
	if achievements[aid] then
		for cid = 1, achievements[aid].numCriteria do
			local cString, cType, completed, quantity, reqQuantity, charName, flags, assetID, quantityString, cid2 = GetAchievementCriteriaInfo(aid, cid)
			achievements[aid].criteria[cid] = new()
			achievements[aid].criteria[cid].name = cString
			achievements[aid].criteria[cid].cType = cType
			achievements[aid].criteria[cid].completed = completed or 0
			achievements[aid].criteria[cid].quantity = quantity
			achievements[aid].criteria[cid].reqQuantity = reqQuantity
		end
		module:updateAchievementTracker()
	end
end

-------------
--- Hooks ---
-------------
function module:AddTrackedAchievement(aid, batch)
	numTrackedAchievement = numTrackedAchievement + 1
	chdb.achievementTracked[aid] = true
	achievements[aid] = new()
	_, achievements[aid].name, _, achievements[aid].completed, _, _, _, achievements[aid].description, _, achievements[aid].image = GetAchievementInfo(aid)
	achievements[aid].numCriteria = GetAchievementNumCriteria(aid)
	achievements[aid].numBooleanCriteria = 0
	achievements[aid].numCompletedBooleanCriteria = 0
	achievements[aid].criteria = new()
	for cid = 1, achievements[aid].numCriteria do
		local cString, cType, completed, quantity, reqQuantity, charName, flags, assetID, quantityString, cid2 = GetAchievementCriteriaInfo(aid, cid)
		achievements[aid].criteria[cid] = new()
		achievements[aid].criteria[cid].name = cString
		achievements[aid].criteria[cid].cType = cType
		achievements[aid].criteria[cid].completed = completed or 0
		achievements[aid].criteria[cid].quantity = quantity
		achievements[aid].criteria[cid].reqQuantity = reqQuantity
		if not reqQuantity or reqQuantity <= 1 then
			achievements[aid].numBooleanCriteria = achievements[aid].numBooleanCriteria + 1
			if achievements[aid].criteria[cid].completed == 1 then
				achievements[aid].numCompletedBooleanCriteria = achievements[aid].numCompletedBooleanCriteria + 1
			end
		end
	end
	if achievements[aid].numBooleanCriteria > 0 then
		booleanCriteriaTracked = booleanCriteriaTracked + 1
		achievements[aid].boolean = true
		if achievements[aid].numBooleanCriteria > maxNumberOfBooleanCriteria then
			maxNumberOfBooleanCriteria = achievements[aid].numBooleanCriteria
		end
		if achievements[aid].numCompletedBooleanCriteria > maxNumberOfCompletedBooleanCriteria then
			maxNumberOfCompletedBooleanCriteria = achievements[aid].numCompletedBooleanCriteria
		end
	end
	local categoryName, subCategoryName
	achievements[aid].subCategory = GetAchievementCategory(aid)
	subCategoryName, achievements[aid].category = GetCategoryInfo(achievements[aid].subCategory)
	categoryName = GetCategoryInfo(achievements[aid].category)
	if achievements[aid].category and achievements[aid].category > 0 then
		if not category[achievements[aid].category] then
			category[achievements[aid].category] = new()
			category[achievements[aid].category].subCategory = new()
			category[achievements[aid].category].numSubCategory = 0
			category[achievements[aid].category].achievements = new()
			category[achievements[aid].category].numAchievements = 0
			category[achievements[aid].category].name = categoryName
			category[achievements[aid].category].numTA, category[achievements[aid].category].numTCA  = GetCategoryNumAchievements(achievements[aid].category)
		end
		if not category[achievements[aid].category].subCategory[achievements[aid].subCategory] then
			category[achievements[aid].category].subCategory[achievements[aid].subCategory] = new()
			category[achievements[aid].category].numSubCategory = category[achievements[aid].category].numSubCategory + 1
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements = 0
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].achievements = new()
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].name = subCategoryName
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].numTA, category[achievements[aid].category].subCategory[achievements[aid].subCategory].numTCA  = GetCategoryNumAchievements(achievements[aid].subCategory)
		end
		category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements = category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements + 1
		category[achievements[aid].category].subCategory[achievements[aid].subCategory].achievements[aid] = true
	elseif achievements[aid].subCategory and achievements[aid].subCategory > 0 then
		if not category[achievements[aid].subCategory] then
			category[achievements[aid].subCategory] = new()
			category[achievements[aid].subCategory].subCategory = new()
			category[achievements[aid].subCategory].numSubCategory = 0
			category[achievements[aid].subCategory].achievements = new()
			category[achievements[aid].subCategory].numAchievements = 0
			category[achievements[aid].subCategory].name = subCategoryName
			category[achievements[aid].subCategory].numTA, category[achievements[aid].subCategory].numTCA  = GetCategoryNumAchievements(achievements[aid].subCategory)
		end
		category[achievements[aid].subCategory].numAchievements = category[achievements[aid].subCategory].numAchievements + 1
		category[achievements[aid].subCategory].achievements[aid] = true
	end
	if not batch then
		if AchievementFrame then
			AchievementFrameAchievements_ForceUpdate()
		end
		module:updateAchievementTracker()
	end
end

function module:RemoveTrackedAchievement(aid)
	if not chdb.achievementTracked[aid] or not achievements[aid] then return end
	numTrackedAchievement = numTrackedAchievement - 1
	if achievements[aid].boolean then
		booleanCriteriaTracked = booleanCriteriaTracked - 1
		if booleanCriteriaTracked == 0 then
			maxNumberOfBooleanCriteria = 0
		else
			maxNumberOfBooleanCriteria = 0
			for aid2, achievementData in pairs(achievements) do
				if aid ~= aid2 and  achievementData.numBooleanCriteria > maxNumberOfBooleanCriteria then
					maxNumberOfBooleanCriteria = achievementData.numBooleanCriteria
				end
			end
		end
	end
	chdb.achievementTracked[aid] = nil
	if achievements[aid].category and achievements[aid].category > 0 and achievements[aid].subCategory and achievements[aid].subCategory > 0 then
		category[achievements[aid].category].subCategory[achievements[aid].subCategory].achievements[aid] = nil
		category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements = category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements - 1
		if category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements == 0 then
			category[achievements[aid].category].subCategory[achievements[aid].subCategory] = del(category[achievements[aid].category].subCategory[achievements[aid].subCategory])
			category[achievements[aid].category].numSubCategory = category[achievements[aid].category].numSubCategory - 1
			if category[achievements[aid].category].numSubCategory == 0 and category[achievements[aid].category].numAchievements == 0 then
				category[achievements[aid].category] = del(category[achievements[aid].category])
			end
		end
	elseif achievements[aid].subCategory and achievements[aid].subCategory > 0 then
		category[achievements[aid].subCategory].achievements[aid] = nil
		category[achievements[aid].subCategory].numAchievements = category[achievements[aid].subCategory].numAchievements - 1
		if category[achievements[aid].subCategory].numSubCategory == 0 and category[achievements[aid].subCategory].numAchievements == 0 then
			category[achievements[aid].subCategory] = del(category[achievements[aid].subCategory])
		end
	end
	achievements[aid] = del(achievements[aid])
	module:updateAchievementTracker()
	if AchievementFrame then
		AchievementFrameAchievements_ForceUpdate()
	end
end

function module:GetNumTrackedAchievements()
	return numTrackedAchievement
end

---------------------------------------------------
--- QTIP Mouse Handler and Custom Cell Provider ---
---------------------------------------------------
local interactivity = {
	chatLink = function(cValue1)
		ChatFrameEditBox:Show()
		ChatEdit_InsertLink(GetAchievementLink(cValue1))
	end,
	achievementTooltip = function(cValue1)
		ItemRefTooltip:ClearLines()
		ShowUIPanel(ItemRefTooltip)
		if not ItemRefTooltip:IsVisible() then
			ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
		end
		local link = GetAchievementLink(cValue1)
		if link then
			ItemRefTooltip:SetHyperlink(link)
		end
	end,
	removeTrackedAchievement = function(cValue1)
		RemoveTrackedAchievement(tonumber(cValue1))
	end,
	displayAchievement = function(cValue1)
		ToggleAchievementFrame()
		if AchievementFrame:IsShown() then
			AchievementFrame_SelectAchievement(tonumber(cValue1))
		end
	end,
	nothing = function(cValue1)
		return
	end,
}

function module:achievementOnMouseDown(cell, button)
	local category, cType, cValue1, cValue2, trackerAnchor = cell.category, cell.cType, cell.cValue1, cell.cValue2, cell.anchor
	if cType == "category" then
		if chdb.categoryDisplay[cValue1] == 1 then
			chdb.categoryDisplay[cValue1] = 0
		else
			chdb.categoryDisplay[cValue1] = 1
		end
		module:updateAchievementTracker()
	elseif cType == "achievement" then
		if button == "RightButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseRightShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseRightControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseRightAlt](cValue1)
			else
				interactivity[prdb.mouseRight](cValue1)
			end
		elseif button == "LeftButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseLeftShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseLeftControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseLeftAlt](cValue1)
			else
				interactivity[prdb.mouseLeft](cValue1)
			end
		elseif button == "MiddleButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseMiddleShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseMiddleControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseMiddleAlt](cValue1)
			else
				interactivity[prdb.mouseMiddle](cValue1)
			end
		end
	end
end

function module:achievementOnEnter(cell)
	local category, cType, cValue1, cValue2, trackerAnchor = cell.category, cell.cType, cell.cValue1, cell.cValue2, cell.anchor
	if cType == "achievement" then
		GameTooltip:SetOwner(usedTracker, trackerAnchor)
		GameTooltip:SetHyperlink(GetAchievementLink(cValue1))
	end
end

function module:achievementOnLeave(cell)
	GameTooltip:Hide()
end

local cProvider, cPrototype, basecCellPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)
function cPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, disableInteraction, category, cType, cValue1, cValue2, trackerAnchor)
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth)
	if not disableInteraction then
		self:EnableMouse(true)
		self.category, self.cType, self.cValue1, self.cValue2, self.anchor = category, cType, cValue1, cValue2, trackerAnchor
		self:SetScript("OnEnter", function(frame) module:achievementOnEnter(frame) end)
		self:SetScript("OnLeave", function(frame) module:achievementOnLeave(frame) end)
		self:SetScript("OnMouseDown", function(frame, button) module:achievementOnMouseDown(frame, button) end)
	else
		self:EnableMouse(false)
		self.category, self.cType, self.cValue1, self.cValue2, self.anchor = nil, nil, nil, nil, nil
		self:SetScript("OnEnter", nil)
		self:SetScript("OnLeave", nil)
		self:SetScript("OnMouseDown", nil)
	end
	return width, height
end

function cPrototype:ReleaseCell()
	self:EnableMouse(false)
	self.category, self.cType, self.cValue1, self.cValue2, self.anchor = nil, nil, nil, nil, nil
	self:SetScript("OnEnter", nil)
	self:SetScript("OnLeave", nil)
	self:SetScript("OnMouseDown", nil)
end

function module:updateAchievementTracker()
	if core:getModuleEnabled("questsTracker") and prdb.showInQuestTracker then
		if achievementTracker then
			QTIP:Release(achievementTracker)
			achievementTracker = nil
		end
		core.cbh:Fire("updateQuestsTracker")
	elseif prdb.toggleTracker and numTrackedAchievement > 0 and not hideInCombat then
		achievementTracker = QTIP:Acquire("tomQuest2Achievements", prdb.numBooleanCriteriaColumns + 1)
		achievementTracker:Clear()
		achievementTracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
		achievementTracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
		achievementTracker:SetScale(prdb.scale)
		achievementTracker:SetFrameStrata("BACKGROUND")
		module:updateAchievement(achievementTracker, trackerAnchor, prdb.disableInteraction, prdb.maxWidth, prdb.statusMaxWidth)
		achievementTracker:ClearAllPoints()
		if prdb.showAnchor then
			achievementTracker:SetPoint(anchorSetPoint[prdb.anchorPosition].point, anchor, anchorSetPoint[prdb.anchorPosition].relPoint)
		else
			achievementTracker:SetPoint(anchorSetPoint[prdb.anchorPosition].point, anchor, anchorSetPoint[prdb.anchorPosition].relPoint, anchorSetPoint[prdb.anchorPosition].x, anchorSetPoint[prdb.anchorPosition].y)
		end
		if not prdb.disableInteraction then
			achievementTracker:UpdateScrolling(prdb.hasMaxHeight and prdb.maxHeight or nil)
			achievementTracker:EnableMouse(true)
		else
			achievementTracker:EnableMouse(false)
		end
		achievementTracker:Show()
	else
		QTIP:Release(usedTracker)
		usedTracker = nil
		achievementTracker = nil
		return
	end
end

function module:displayAchievement()
	if numTrackedAchievement == 0 then 
		return false, 2
	else
		return true, prdb.numBooleanCriteriaColumns + 1
	end
end

local function addAchievementToTracker(aid, tracker, trackerAnchor, disableInteraction, maxWidth, statusMaxWidth)
	tempLine = tracker:AddLine()
	local achImage = prdb.displayAchievementImage and format("|T%s:%d|t ", achievements[aid].image, prdb.achievementImageSize) or ""
	tracker:SetCell(tempLine, 1, format("%s|cffffffff%s|r", achImage, achievements[aid].name), aATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, prdb.achievementCellPadding, 0, maxWidth, maxWidth, disableInteraction, "achievements", "achievement", aid, nil, trackerAnchor)
	tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, " ", acATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, statusMaxWidth, statusMaxWidth, disableInteraction, "achievements", "achievement", aid, nil, trackerAnchor)
	if achievements[aid].numCriteria and achievements[aid].numCriteria > 0 then
		local cIndex = 1
		for cid, acData in pairs (achievements[aid].criteria) do
			if (prdb.hideCompletedCriteria and acData.completed == 0) or (not prdb.hideCompletedCriteria) then
				if acData.reqQuantity > 1 then
					if (acData.cType == 7 and acData.quantity > 0) or acData.cType ~= 7 then
						tempLine = tracker:AddLine()
						local criteriacValue1 = acData.cType == 62 and formatMoney(acData.quantity) or (acData.cType == 67 and formatMoney(acData.quantity) or format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(acData.quantity / acData.reqQuantity), acData.quantity, acData.reqQuantity))
						tracker:SetCell(tempLine, 1, format("|cff%s%s|r", core:getColorFromCompletionHEX(acData.quantity / acData.reqQuantity), acData.name), acATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, prdb.criteriaCellPadding, 0, maxWidth, maxWidth, disableInteraction, "achievements", "achievement", aid, nil, trackerAnchor)
						tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, criteriacValue1, acATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, statusMaxWidth, statusMaxWidth, disableInteraction, "achievements", "achievement", aid, nil, trackerAnchor)
					end
				else
					if cIndex == 1 then	
						tempLine = tracker:AddLine() 
						tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, " ", acATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, statusMaxWidth, statusMaxWidth, disableInteraction, "achievements", "achievement", aid, nil, trackerAnchor)
					end
					tracker:SetCell(tempLine, cIndex, format("|cff%s%s|r", core:getColorFromCompletionHEX(acData.quantity), acData.name), acATrFontTQ2, "LEFT", 1, cProvider, cIndex == 1 and prdb.criteriaCellPadding or 0, 0, columnWidth, columnWidth, disableInteraction, "achievements", "achievement", aid, nil, trackerAnchor)
					if cIndex == prdb.numBooleanCriteriaColumns then cIndex = 1 else cIndex = cIndex + 1 end
				end
			end
		end
	else
		tempLine = tracker:AddLine()
		tracker:SetCell(tempLine, 1, achievements[aid].description, acATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, prdb.criteriaCellPadding, 0, maxWidth, maxWidth, disableInteraction, "achievements", "achievement", aid, nil, trackerAnchor)
		tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, " ", acATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, statusMaxWidth, statusMaxWidth, disableInteraction, "achievements", "achievement", aid, nil, trackerAnchor)
	end
end

function module:updateAchievement(tracker, trackerAnchor, disableInteraction, maxWidth, statusMaxWidth)
	usedTracker = tracker
	local tempLine
	local columnWidth = maxWidth/prdb.numBooleanCriteriaColumns
	tempLine = tracker:AddLine()
	for i = 1, prdb.numBooleanCriteriaColumns - 1 do
		tracker:SetCell(tempLine, i, " ", acATrFontTQ2, "LEFT", 1, cProvider, 0, 0, columnWidth, columnWidth, disableInteraction, "inactive")
	end
	tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns, " ", acATrFontTQ2, "RIGHT", 1, cProvider, 0, 0, statusMaxWidth, statusMaxWidth, disableInteraction, "inactive")
	if prdb.showByCategory then
		for cid, cData in pairs(category) do
			chdb.categoryDisplay[cid] = chdb.categoryDisplay[cid] or 1
			tempLine = tracker:AddLine()
			tracker:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r", chdb.categoryDisplay[cid] == 0 and "Plus" or "Minus", prdb.categoryCollapseButtonSize, core:getColorFromCompletionHEX(cData.numTCA / cData.numTA) or "ffffff", cData.name), cATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, 0, 0, maxWidth, maxWidth, disableInteraction, "achievements", "category", cid, nil, trackerAnchor)
			tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(cData.numTCA / cData.numTA), cData.numTCA , cData.numTA), cATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, statusMaxWidth, statusMaxWidth, disableInteraction, "achievements", "category", cid, nil, trackerAnchor)
			if chdb.categoryDisplay[cid] == 1 then
				if cData.numAchievements > 0 then
					for aid in pairs(cData.achievements) do
						addAchievementToTracker(aid, tracker, trackerAnchor, disableInteraction, maxWidth, statusMaxWidth)
					end
				end
				if cData.numSubCategory > 0 then
					for scid, scData in pairs(cData.subCategory) do
						chdb.categoryDisplay[scid] = chdb.categoryDisplay[scid] or 1
						tempLine = tracker:AddLine()
						tracker:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r", chdb.categoryDisplay[scid] == 0 and "Plus" or "Minus", prdb.subCategoryCollapseButtonSize, core:getColorFromCompletionHEX(scData.numTCA / scData.numTA), scData.name), scATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, prdb.subCategoryCellPadding, 0, maxWidth, maxWidth, disableInteraction, "achievements", "category", scid, nil, trackerAnchor)
						tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(scData.numTCA / scData.numTA), scData.numTCA , scData.numTA), scATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, statusMaxWidth, statusMaxWidth, disableInteraction, "achievements", "category", scid, nil, trackerAnchor)
						if chdb.categoryDisplay[scid] == 1 then
							if scData.numAchievements > 0 then
								for aid in pairs(scData.achievements) do
									addAchievementToTracker(aid, tracker, trackerAnchor, disableInteraction, maxWidth, statusMaxWidth)
								end
							end
						end
					end
				end
			end
		end
	else
		for aid, aData in pairs(achievements) do
			addAchievementToTracker(aid, tracker, trackerAnchor, disableInteraction, maxWidth, statusMaxWidth)
		end
	end
end

--------------------------
--- Keybinding globals ---
--------------------------
BINDING_NAME_TQ2ACHIEVEMENTTRACKERTOGGLE = LATr["Achievement Tracker"]

function tQ2toggleAchievementTracker()
	prdb.toggleTracker = not prdb.toggleTracker
	if not prdb.toggleTracker then
		QTIP:Release(usedTracker)
		usedTracker = nil
		achievementTracker = nil
	else
		module:updateAchievementTracker()
	end
end