-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("questsTooltip", "AceEvent-3.0")

-----------------------
--- Library loading ---
-----------------------
local LQTo = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-QuestsTooltip")
local LSM = LibStub("LibSharedMedia-3.0")
local QTIP = LibStub("LibQTip-1.0")
local dataobj = LibStub("LibDataBroker-1.1"):NewDataObject("tomQuest2", {type = "data source", icon = "Interface\\AddOns\\tomQuest2\\media\\icon"})

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs, unpack, tonumber = pairs, unpack, tonumber
local format, sub = string.format, string.sub
local ToggleFrame = ToggleFrame
local ShowUIPanel = ShowUIPanel
local QuestLogPushQuest = QuestLogPushQuest
local MouseIsOver = MouseIsOver
local ItemRefTooltip = ItemRefTooltip
local ChatEdit_InsertLink = ChatEdit_InsertLink
local SelectQuestLogEntry, GetQuestLogSelection = SelectQuestLogEntry, GetQuestLogSelection
local GetQuestLink = GetQuestLink
local CreateFrame, CreateFont, SetFont = CreateFrame, CreateFont, SetFont
local IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown = IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown
local GetAbandonQuestItems, GetAbandonQuestName, SetAbandonQuest = GetAbandonQuestItems, GetAbandonQuestName, SetAbandonQuest

-----------------
--- Variables ---
-----------------
module.modName = LQTo["Broker Quests Tooltip"]
local _
local data = {}
local prdb, chdb, options
local tip, tipAnchor, anchor
local tooltipIsShown = false
local maxQuests = MAX_QUESTS
local timer = CreateFrame("Frame")
local hToFontTQ2 = CreateFont("hToFontTQ2")
local qToFontTQ2 = CreateFont("qToFontTQ2")
local qoToFontTQ2 = CreateFont("qoToFontTQ2")
local outlines = {[""] = LQTo["None"], ["OUTLINE"] = LQTo["Outline"], ["THICKOUTLINE"] = LQTo["Thick Outline"]}
local hideObjectives = {[1] = LQTo["When you've completed the objective"], [2] = LQTo["When everyone in the party has completed the objective"], [3] = LQTo["Never"]}
local defaultColors = {"defaultHeaderNameColor", "defaultHeaderLevelColor", "defaultQuestNameColor", "defaultQuestLevelColor", "defaultQuestObjectiveColor"}
local interactivityList = {
	["chatLink"] = LQTo["Insert quest link in chat"],
	["objectiveProgress"] = LQTo["Insert quest objectives progress in chat"],
	["macroUpdate"] = LQTo["Update special quest item macro"],
	["questTooltip"] = LQTo["Display of quest tooltip"],
	["shareQuest"] = LQTo["Share quest"],
	["trackQuest"] = LQTo["Track quest"],
	["abandonQuest"] = LQTo["Abandon quest"],
	["displayQuest"] = LQTo["Diplay quest log"],
	["displayQuestInfo"] = LQTo["Display quest on map (external module needed)"],
	["displayCQIStartNPC"] = LQTo["Display of start npc in Cartographer quest Info"],
	["displayCQIEndNPC"] = LQTo["Display of end npc in Cartographer Quest Info"],
	["displayCQIObjective"] = LQTo["Display of objective in Cartographer Quest Info"],
	["displayMMObjective"] = LQTo["Display of objective in your map with MobMap"],
	["nothing"] = LQTo["Do nothing"],
}
local ldbInteractivityList = {
	["openOptions"] = LQTo["Open tomQuest2 option frame"],
	["toggleQuestTracker"] = LQTo["Toggle quests tracking frame"],
	["toggleAchievementTracker"] = LQTo["Toggle achievements tracking frame"],
	["toggleAnnouncementModule"] = LQTo["toggle chat announcement module"],
	["toggleQuestsLog"] = LQTo["Toggle Blizzard quests log"],
	["toggleSyncPartyModule"] = LQTo["Toggle party synchronization module"],
	["toggleQuestsTrackerModule"] = LQTo["Toggle quests tracking module"],
	["toggleAchievementTrackerModule"] = LQTo["Toggle achievements tracking module"],
	["nothing"] = LQTo["Do nothing"],
}
local interactivityAction = {
	mouseRightShift = LQTo["Shift + Right click"],
	mouseRightControl = LQTo["Ctrl + Right click"],
	mouseRightAlt = LQTo["Alt + Right click"],
	mouseRight = LQTo["Right click"],
	mouseLeftShift = LQTo["Shift + Left click"],
	mouseLeftControl = LQTo["Ctrl + Left click"],
	mouseLeftAlt = LQTo["Alt + Left click"],
	mouseLeft = LQTo["Left click"],
	mouseMiddleShift = LQTo["Shift + Middle click"],
	mouseMiddleControl = LQTo["Ctrl + Middle click"],
	mouseMiddleAlt = LQTo["Alt + Middle click"],
	mouseMiddle = LQTo["Middle click"],
}

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		showNumCompletedQuests = true,
		showNumQuests = true,
		showMaxQuest = true,
		hideCompletedObjective = 2,
		headerFont = LSM.DefaultMedia.font,
		headerFontSize = 12,
		headerFontOutline = "", 
		questFont = LSM.DefaultMedia.font,
		questFontSize = 10,
		questFontOutline = "",
		questObjectiveFont = LSM.DefaultMedia.font,
		questObjectiveFontSize = 8, 
		questObjectiveFontOutline = "",
		maxWidth = 300,
		statusMaxWidth = 100,
		hasMaxHeight = true,
		maxHeight = 600,
		scale = 1,
		showQuestItemIcon = true,
		questItemIconSize = 24,
		questsCellPadding = 5,
		objectivesCellPadding = 10,
		rightCellPadding = 5,
		collapseButtonSize = 24,
		colorizeHeaderName = true,
		defaultHeaderNameColor = "ffffff",
		hideZoneLevelRange = false,
		colorizeHeaderLevel = true,
		defaultHeaderLevelColor = "ffffff",
		colorizeQuestName = true,
		defaultQuestNameColor = "ffffff",
		colorizeQuestLevel = true,
		defaultQuestLevelColor = "ffffff",
		colorizeQuestObjective = true,
		defaultQuestObjectiveColor = "ffffff",
		colorizeQuestObjectiveCounter = true,
		showHints = true,
		overTooltip = true,
		mouseRightShift = "chatLink",
		mouseRightControl = "objectiveProgress",
		mouseRightAlt = "macroUpdate",
		mouseRight = "questTooltip",
		mouseLeftShift = "trackQuest",
		mouseLeftControl = "shareQuest",
		mouseLeftAlt = "abandonQuest",
		mouseLeft = "displayQuest",
		mouseMiddleShift = "nothing",
		mouseMiddleControl = "nothing",
		mouseMiddleAlt = "nothing",
		mouseMiddle = "nothing",
		ldbMouseRightShift = "openOptions",
		ldbMouseRightControl = "openOptions",
		ldbMouseRightAlt = "openOptions",
		ldbMouseRight = "openOptions",
		ldbMouseLeftShift = "toggleQuestTracker",
		ldbMouseLeftControl = "toggleAnnouncementModule",
		ldbMouseLeftAlt = "toggleAchievementTracker",
		ldbMouseLeft = "toggleQuestsLog",
		ldbMouseMiddleShift = "toggleQuestTracker",
		ldbMouseMiddleControl = "toggleAnnouncementModule",
		ldbMouseMiddleAlt = "toggleAchievementTracker",
		ldbMouseMiddle = "toggleQuestsLog",
	}
}

---------------
--- Options ---
---------------
local optGetter, optSetter, colorGetter, colorSetter, hexColorGetter, hexColorSetter, ldbTextSetter, headFontSetter, questFontSetter, qoFontSetter
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
	end
	function colorGetter(info)
		return unpack(prdb[info[#info]])
	end
	function colorSetter(info, arg1, arg2, arg3, arg4)
		prdb[info[#info]][1] = arg1
		prdb[info[#info]][2] = arg2
		prdb[info[#info]][3] = arg3
		prdb[info[#info]][4] = arg4
	end
	function hexColorGetter(info)
		return tonumber(sub(prdb[info[#info]], 1, 2), 16) / 255, tonumber(sub(prdb[info[#info]], 3, 4), 16) / 255, tonumber(sub(prdb[info[#info]], 5, 6), 16) / 255
	end
	function hexColorSetter(info, arg1, arg2, arg3)
		prdb[info[#info]] = format("%02x%02x%02x", arg1*255, arg2*255, arg3*255)
	end
	function ldbTextSetter(info, value)
		prdb[info[#info]] = value
		module:questLogUpdated(nil, true, false)
	end
	function headFontSetter(info, value)
		prdb[info[#info]] = value
		hToFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	end
	function questFontSetter(info, value)
		prdb[info[#info]] = value
		qToFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	end
	function qoFontSetter(info, value)
		prdb[info[#info]] = value
		qoToFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			childGroups = "tab",
			name = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				questsCounter = {
					type = "group",
					name = LQTo["Quests counter"],
					set = ldbTextSetter,
					args = {
						showNumCompletedQuests = {
							name = LQTo["Show number of completed quest"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						showNumQuests = {
							name = LQTo["Show number of quests"],
							type = "toggle",
							width = "full",
							order = 2,
						},
						showMaxQuest = {
							name = LQTo["Show maximum quests number"],
							type = "toggle",
							width = "full",
							order = 3,
						},
					},
				},
				display	= {
					type = "group",
					name = LQTo["Display"],
					args = {
						hideCompletedObjective = {
							name = LQTo["Hide completed quests Objectives"],
							type = "select",
							values = hideObjectives,
							width = "full",
							order = 1,
						},
						hideZoneLevelRange = {
							name = LQTo["Hide zone level range"],
							type = "toggle",
							width = "full",
							order = 2,
						},
					},
				},
				fonts = {
					type = "group",
					name = LQTo["Tooltip fonts"],
					args = {
						zone = {
							type = "group",
							name = LQTo["Zone"],
							set = headFontSetter,
							inline = true,
							order = 1,
							args = {
								headerFont = {
									name = LQTo["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								headerFontSize = {
									name = LQTo["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								headerFontOutline = {
									name = LQTo["Font outline"],
									type = "select",
									values = outlines,
									width = "full",
									order = 3,
								},
							},
						}, 
						quests = {
							type = "group",
							name = LQTo["Quests"],
							set = questFontSetter,
							inline = true,
							order = 2,
							args = {
								questFont = {
									name = LQTo["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								questFontSize = {
									name = LQTo["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								questFontOutline = {
									name = LQTo["Font outline"],
									type = "select",
									values = outlines,
									width = "full",
									order = 3,
								},
							},
						},
						questsObjectives = {
							type = "group",
							name = LQTo["Quests objectives"],
							set = qoFontSetter,
							inline = true,
							order = 3,
							args = {
								questObjectiveFont = {
									name = LQTo["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								questObjectiveFontSize = { 
									name = LQTo["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								questObjectiveFontOutline = {
									name = LQTo["Font outline"],
									type = "select",
									values = outlines,
									width = "full",
									order = 3,
								},
							},
						},
					},
				},
				appearance = {
					type = "group",
					name = LQTo["Tooltip appearance"],
					args = {
						scale = {
							name = LQTo["Quests Tooltip scale"],
							type="range",
							min = 0,
							max = 1,
							step = 0.1,
							width = "full",
							order = 1,
						},
						hasMaxHeight = {
							name = LQTo["Set quests tooltip maximum height after which it will be scrollable"],
							type = "toggle",
							width = "full",
							order = 2,
						},
						maxHeight = {
							name = LQTo["Quests tooltip max height"],
							type = "range",
							min = 100,
							max = 2000,
							step = 5,
							width = "full",
							order = 3,
						},
						collapseButtonSize = {
							name = LQTo["Collapse Button Size"],
							type = "range",
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 4,
						},
						questsCellPadding = {
							name = LQTo["Quests cell padding"],
							type="range",
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 5,
						},
						objectivesCellPadding = {
							name = LQTo["Quests objectives cell padding"],
							type="range",
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 6,
						},
						rightCellPadding = {
							name = LQTo["Right-most cell right padding to avoid cell clipping"],
							type = "range",
							min = 0,
							max = 10,
							step = 1,
							width = "full",
							order = 7,
						},
						maxWidth = {
							name = LQTo["Quests Tooltip column max width"],
							type="range",
							min = 50,
							max = 600,
							step = 5,
							width = "full",
							order = 8,
						},
						statusMaxWidth = {
							name = LQTo["Quests Tooltip status column max width"],
							type="range",
							min = 30,
							max = 200,
							step = 5,
							width = "full",
							order = 9,
						},
					},
				},
				colors = {
					type = "group",
					name = LQTo["Colors"],
					args = {
						colorizeHeaderName = {
							name = LQTo["Colorize Quests Header Name"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						defaultHeaderNameColor = {
							name = LQTo["Default header name color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 2,
						},
						colorizeHeaderLevel = {
							name = LQTo["Colorize Quests Header Level"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						defaultHeaderLevelColor = {
							name = LQTo["Default header level color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 4,
						},
						colorizeQuestName = {
							name = LQTo["Colorize Quests Name"],
							type = "toggle",
							width = "full",
							order = 5,
						},
						defaultQuestNameColor = {
							name = LQTo["Default quest name color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 6,
						},
						colorizeQuestLevel = {
							name = LQTo["Colorize Quests Level"],
							type = "toggle",
							width = "full",
							order = 7,
						},
						defaultQuestLevelColor = {
							name = LQTo["Default quest level color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 8,
						},
						colorizeQuestObjective = {
							name = LQTo["Colorize Quests Objectives"],
							type = "toggle",
							width = "full",
							order = 9,
						},
						defaultQuestObjectiveColor = {
							name = LQTo["Default quest objective color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 10,
						},
						colorizeQuestObjectiveCounter = {
							name = LQTo["Colorize quests objectives counter"],
							type = "toggle",
							width = "full",
							order = 11,
						},
					},
				},
				questsItems = {
					type = "group",
					name = LQTo["Quests items"],
					args = {
						showQuestItemIcon = {
							name = LQTo["Show special quest item icon behind quest name"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						questItemIconSize = {
							name = LQTo["Quest item icon size"],
							type = "range",
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 2,
						},
					},
				},
				tooltipInteraction = {
					type = "group",
					name = LQTo["Tooltip interaction"],
					args = {
						showHints = {
							name = LQTo["Show interractivity hints at the bottom of the tooltip"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						overTooltip = {
							name = LQTo["Display quest tooltip on mouse over"],
							type = "toggle",
							width = "full",
							order = 2,
						},
						mouseRightShift = {
							name = LQTo["Action on right shift click"],
							type = "select",
							values = interactivityList,
							width = "full",
							order = 3,
						},
						mouseRightControl = {
							name = LQTo["Action on right control click"],
							type = "select",
							values = interactivityList,
							width = "full",
							order = 4,
						},
						mouseRightAlt = {
							name = LQTo["Action on right alt click"],
							type = "select",
							values = interactivityList,
							width = "full",
							order = 5,
						},
						mouseRight = {
							name = LQTo["Action on right click"],
							type = "select",
							values = interactivityList,
							width = "full",
							order = 6,
						},
						mouseLeftShift = {
							name = LQTo["Action on left shift click"],
							type = "select",
							values = interactivityList,
							width = "full",
							order = 7,
						},
						mouseLeftControl = {
							name = LQTo["Action on left control click"],
							type = "select",
							values = interactivityList,
							width = "full",
							order = 8,
						},
						mouseLeftAlt = {
							name = LQTo["Action on left alt click"],
							type = "select",
							values = interactivityList,
							width = "full",
							order = 9,
						},
						mouseLeft = {
							name = LQTo["Action on left click"],
							type = "select",
							values = interactivityList,
							width = "full",
							order = 10,
						},
						mouseMiddleShift = {
							name = LQTo["Action on middle shift click"],
							type = "select",
							values = interactivityList,
							width = "full",
							order = 11,
						},
						mouseMiddleControl = {
							name = LQTo["Action on middle control click"],
							type = "select",
							values = interactivityList,
							width = "full",
							order = 12,
						},
						mouseMiddleAlt = {
							name = LQTo["Action on middle alt click"],
							type = "select",
							values = interactivityList,
							width = "full",
							order = 13,
						},
						mouseMiddle = {
							name = LQTo["Action on middle click"],
							type = "select",
							values = interactivityList,
							width = "full",
							order = 14,
						},
					},
				},
				ldbInteraction = {
					type = "group",
					name = LQTo["Quests counter interaction"],
					args = {
						ldbMouseRightShift = {
							name = LQTo["Action on right shift click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 1,
						},
						ldbMouseRightControl = {
							name = LQTo["Action on right control click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 2,
						},
						ldbMouseRightAlt = {
							name = LQTo["Action on right alt click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 3,
						},
						ldbMouseRight = {
							name = LQTo["Action on right click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 4,
						},
						ldbMouseLeftShift = {
							name = LQTo["Action on left shift click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 5,
						},
						ldbMouseLeftControl = {
							name = LQTo["Action on left control click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 6,
						},
						ldbMouseLeftAlt = {
							name = LQTo["Action on left alt click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 7,
						},
						ldbMouseLeft = {
							name = LQTo["Action on left click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 8,
						},
						ldbMouseMiddleShift = {
							name = LQTo["Action on middle shift click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 9,
						},
						ldbMouseMiddleControl = {
							name = LQTo["Action on middle control click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 10,
						},
						ldbMouseMiddleAlt = {
							name = LQTo["Action on middle alt click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 11,
						},
						ldbMouseMiddle = {
							name = LQTo["Action on middle click"],
							type = "select",
							values = ldbInteractivityList,
							width = "full",
							order = 12,
						},
					},
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("questsTooltip", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	self:SetEnabledState(core:getModuleEnabled("questsTooltip"))
	core:registerModuleOptions("questsTooltip", getOptions, module.modName)
	chdb.collapseAll = chdb.collapseAll or 0
	chdb.zoneDisplay = chdb.zoneDisplay or {}
	core.RegisterCallback(self, "updateDBtoV3")
	LSM.RegisterCallback(self, "LibSharedMedia_Registered", "updateUsedMedia")
end

function module:updateDBtoV3(event)
	for k in pairs(defaultColors) do
		if type(prdb[k]) == "table" then
			prdb[k] = format("%02x%02x%02x", (prdb[k][1] or 0)*255, (prdb[k][2] or 0)*255, (prdb[k][3] or 0)*255)
		end
	end
end

local function hideTooltip()
	if not tip or MouseIsOver(tip) then return end
	tooltipIsShown = false
	tip:SetScript("OnUpdate", nil)
	QTIP:Release(tip)
	tip = nil
end

function module:OnEnable()
	core.RegisterCallback(self, "questLogUpdated")
	hToFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	qToFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	qoToFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
	chdb = self.db.char
	if not self:IsEnabled() then return end
	hToFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	qToFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	qoToFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
end

function module:updateUsedMedia(event, mediatype, key)
	if mediatype == "font" then
		local updateTooltip = nil
		if key == prdb.headerFont then
			updateTooltip = true
			hToFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
		end
		if key == prdb.questFont then
			updateTooltip = true
			qToFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
		end
		if key == prdb.questObjectiveFont then
			updateTooltip = true
			qoToFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
		end
		if tooltipIsShown and updateTooltip then
			module:updateTooltip()
		end
	end
end

---------------------------
--- Main Module Program ---
---------------------------
function module:questLogUpdated(event, ldbText, tooltip)
	if ldbText then
		if prdb.showNumCompletedQuests then
			if prdb.showNumQuests then
				if prdb.showMaxQuest then
					if core.numQuests > 0 then
						dataobj.text = format("|cff%s%s/%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedQuests/core.numQuests), core.numCompletedQuests, core.numQuests, maxQuests)
					else
						dataobj.text = format("0/0/%s", maxQuests)
					end
				else
					if core.numQuests > 0 then
						dataobj.text = format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedQuests/core.numQuests), core.numCompletedQuests, core.numQuests)
					else
						dataobj.text = "0/0"
					end
				end
			elseif prdb.showMaxQuest then
				dataobj.text = format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedQuests/maxQuests), core.numCompletedQuests, maxQuests)
			else
				dataobj.text = core.numCompletedQuests
			end
		elseif prdb.showNumQuests then
			if prdb.showMaxQuest then
				dataobj.text = format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numQuests/maxQuests), core.numQuests, maxQuests)
			else
				dataobj.text = core.numQuests
			end	
		elseif prdb.showMaxQuest then
			dataobj.text = maxQuests
		else
			dataobj.text = ""
		end
	end
	if tooltip and tooltipIsShown then
		module:updateTooltip()
	end
end

---------------------------------------------------
--- QTIP Mouse Handler and Custom Cell Provider ---
---------------------------------------------------
local interactivity = {
	chatLink = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			ChatFrameEditBox:Show()
			ChatEdit_InsertLink(GetQuestLink(core.quests[cValue1].id))
		end
	end,
	objectiveProgress = function(cValue1)
		if core.quests[cValue1].numObjectives > 0 and core.quests[cValue1].complete[core.playerName] then
			ChatFrameEditBox:Show()
			for objective, qoData in pairs(core.questsObjectives[cValue1]) do
				local got, _ = qoData.got[core.playerName]:match("^(.-)~(%d-)$")
				ChatFrameEditBox:Insert(format("{%s %s/%s} ", qoData.desc, got, qoData.need))
			end
		end
	end,
	macroUpdate = function(cValue1)
		core:generateSpecialItemMacro(cValue1)
	end,
	questTooltip = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			ItemRefTooltip:ClearLines()
			ShowUIPanel(ItemRefTooltip)
			if not ItemRefTooltip:IsVisible() then
				ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
			end
			local link = GetQuestLink(core.quests[cValue1].id)
			if link then
				ItemRefTooltip:SetHyperlink(link)
			end
		end
	end,
	shareQuest = function(cValue1)
		if core.numPartyMembers > 0 and core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id and core.quests[cValue1].sharable then
			local wasSelected = GetQuestLogSelection()
			SelectQuestLogEntry(core.quests[cValue1].id)
			QuestLogPushQuest()
			SelectQuestLogEntry(wasSelected)
		end
	end,
	trackQuest = function(cValue1)
		core.cbh:Fire("toggleTrackedQuest", cValue1)
		module:updateTooltip()
	end,
	abandonQuest = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			local wasSelected = GetQuestLogSelection()
			SelectQuestLogEntry(core.quests[cValue1].id)
			SetAbandonQuest()
			local items = GetAbandonQuestItems()
			if items then
				StaticPopup_Hide("ABANDON_QUEST")
				StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
			else
				StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
				StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
			end
			SelectQuestLogEntry(wasSelected)
		end
	end,
	displayQuest = function(cValue1)
		core:displayQuestInformations(cValue1)
	end,
	displayQuestInfo = function(cValue1, cValue2)
		if cValue2 then
			core.cbh:Fire("toggleQuestInfo", cValue1, cValue2)
		end
	end,
	displayCQIStartNPC = function(cValue1)
		if Cartographer_QuestInfo then
			local q = Cartographer_QuestInfo:GetQuest(cValue1)
			if q and q.start_npc then
				data.quest = q.title_full
				data.obj = nil
				data.npc = q.start_npc
				Cartographer_QuestInfo:OnButtonClick(nil, "start", "LeftButton", data)
			end
		end
	end,
	displayCQIEndNPC = function(cValue1)
		if Cartographer_QuestInfo then
			local q = Cartographer_QuestInfo:GetQuest(cValue1)
			if q and q.end_npc then
				data.quest = q.title_full
				data.obj = nil
				data.npc = q.end_npc
				Cartographer_QuestInfo:OnButtonClick(nil, "end", "LeftButton", data)
			end
		end
	end,
	displayCQIObjective = function(cValue1, cValue2)
		if cValue2 and Cartographer_QuestInfo then
			local q = Cartographer_QuestInfo:GetQuest(cValue1)
			if q and q.objs then
				for _, obj in pairs(q.objs) do
					if obj.title:find(core.questsObjectives[cValue1][cValue2].desc, 1, true) then
						data.quest = q.title_full
						data.obj = obj
						data.npc = nil
					end
				end
				if data.obj ~= nil then
					Cartographer_QuestInfo:OnButtonClick(nil, "obj", "LeftButton", data)
				end
			end
		end
	end,
	displayMMObjective = function(cValue1, cValue2)
		if cValue2 and MobMapDotParentFrame and MobMap_ParseQuestObjective then
			MobMap_ParseQuestObjective(core.questsObjectives[cValue1][cValue2].desc)
		end
	end,
	nothing = function(cValue1)
		return
	end,
}

local tipOnMouseDown = function(cell, button)
	local category, cType, cValue1, cValue2, anchor = cell.category, cell.cType, cell.cValue1, cell.cValue2, cell.anchor
	if cType == "header" then
		if chdb.collapseAll == 1 then
			chdb.collapseAll = 0
			for zone in pairs(chdb.zoneDisplay) do
				chdb.zoneDisplay[zone] = 0
			end
		else
			chdb.collapseAll = 1
			for zone in pairs(chdb.zoneDisplay) do
				chdb.zoneDisplay[zone] = 1
			end
		end
		module:updateTooltip()
	elseif cType == "zone" then
		if chdb.zoneDisplay[cValue1] == 1 then
			chdb.zoneDisplay[cValue1] = 0
			chdb.collapseAll = 0
		else
			chdb.zoneDisplay[cValue1] = 1
			chdb.collapseAll = 1
		end
		module:updateTooltip()
	elseif cType == "quest" then
		if button == "RightButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseRightShift](cValue1, cValue2)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseRightControl](cValue1, cValue2)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseRightAlt](cValue1, cValue2)
			else
				interactivity[prdb.mouseRight](cValue1, cValue2)
			end
		elseif button == "LeftButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseLeftShift](cValue1, cValue2)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseLeftControl](cValue1, cValue2)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseLeftAlt](cValue1, cValue2)
			else
				interactivity[prdb.mouseLeft](cValue1, cValue2)
			end
		elseif button == "MiddleButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseMiddleShift](cValue1, cValue2)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseMiddleControl](cValue1, cValue2)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseMiddleAlt](cValue1, cValue2)
			else
				interactivity[prdb.mouseMiddle](cValue1, cValue2)
			end
		end
	end
end

local tipOnEnter = function(cell)
	local category, cType, cValue1, cValue2, anchor = cell.category, cell.cType, cell.cValue1, cell.cValue2, cell.anchor
	if prdb.overTooltip and cType == "quest" then
		if cValue1 and cValue2 and core.questsObjectives[cValue1] and core.questsObjectives[cValue1][cValue2].cType == "object" and core.questsItems[core.questsObjectives[cValue1][cValue2].desc].mobDrop then
			GameTooltip:SetOwner(tip, anchor)
			GameTooltip:AddLine(LQTo["|cffffffffDrop from:|r"])
			for mobName in pairs(core.questsItems[core.questsObjectives[cValue1][cValue2].desc].mobDrop) do
				local zones = ""
				for zone in pairs(core.questsMobsItems[mobName].mobLoc) do
					zones = zones == "" and zone or format("%s, %s", zones, zone)
				end
				if core.questsMobsItems[mobName].items[core.questsObjectives[cValue1][cValue2].desc].dropRate then
					GameTooltip:AddDoubleLine(format(LQTo["  |cffffffff%s in|r |cff00ff00%s|r"], mobName, zones), format("|cffffffff%s%%|r", core.questsMobsItems[mobName].items[core.questsObjectives[cValue1][cValue2].desc].dropRate))
				else
					GameTooltip:AddLine(format(LQTo["  |cffffffff%s in|r |cff00ff00%s|r"], mobName, zones))
				end
			end
			GameTooltip:Show()
		elseif core.quests[cValue1] and core.quests[cValue1].id then
			local link = GetQuestLink(core.quests[cValue1].id)
			if link then
				GameTooltip:SetOwner(tip, anchor)
				GameTooltip:SetHyperlink(link)
			end
		end
	end
end

local tipOnLeave = function(cell)
	GameTooltip:Hide()
end

local cProvider, cPrototype, basecCellPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)
function cPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, category, cType, cValue1, cValue2, anchor)
	self.category, self.cType, self.cValue1, self.cValue2, self.anchor = category, cType, cValue1, cValue2, anchor
	self:EnableMouse(true)
	self:SetScript("OnEnter", tipOnEnter)
	self:SetScript("OnLeave", tipOnLeave)
	self:SetScript("OnMouseDown", tipOnMouseDown)
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth)
	return width, height
end

function cPrototype:ReleaseCell()
	self.category, self.cType, self.cValue1, self.cValue2, self.anchor = nil, nil, nil, nil, nil
	self:EnableMouse(false)
	self:SetScript("OnEnter", nil)
	self:SetScript("OnLeave", nil)
	self:SetScript("OnMouseDown", nil)
end

function module:updateTooltip()
	tip = QTIP:Acquire("tomQuest2ToolTip", 2, "LEFT", "RIGHT")
	tip:Clear()
	tooltipIsShown = true
	tip:SetScale(prdb.scale)
	local tempLine
	tempLine = tip:AddLine()
	tip:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cffffffff%s|r", chdb.collapseAll == 0 and "Plus" or "Minus", prdb.collapseButtonSize,"tomQuest2"), hToFontTQ2, "CENTER", 2, cProvider, 0, 0, prdb.maxWidth + prdb.statusMaxWidth, prdb.maxWidth + prdb.statusMaxWidth, nil, "header", nil, nil, tipAnchor)
	tip:AddLine(" ")
	for _, zone in pairs(core.sortedZones) do
		chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r |cff%s%s|r", chdb.zoneDisplay[zone] == 0 and "Plus" or "Minus", prdb.collapseButtonSize, prdb.colorizeHeaderName and core.zones[zone].color or prdb.defaultHeaderNameColor, zone, prdb.colorizeHeaderLevel and core.zones[zone].color or prdb.defaultHeaderLevelColor, prdb.hideZoneLevelRange and "" or core.zones[zone].level), hToFontTQ2, "LEFT", 2, cProvider, 0, prdb.rightCellPadding, prdb.maxWidth + prdb.statusMaxWidth, prdb.maxWidth + prdb.statusMaxWidth, nil, "zone", zone, nil, tipAnchor)
		if chdb.zoneDisplay[zone] == 1 then
			for _, uid in pairs(core.sortedQuestsByZone[zone]) do
				tempLine = tip:AddLine()
				local questIcon = prdb.showQuestItemIcon and (core.quests[uid].specialItemIcon and format(" |T%s:%d|t", core.quests[uid].specialItemIcon, prdb.questItemIconSize) or "") or ""
				local trackIcon = core.quests[uid].tracked and format(" |TInterface\\Buttons\\UI-CheckBox-Check:%d|t", prdb.questItemIconSize) or ""
				local numPlayerTag = core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""
				tip:SetCell(tempLine, 1, format("|cff%s%s|r |cff%s%s|r%s%s %s ", prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor, core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor, core.quests[uid].title, questIcon, trackIcon, numPlayerTag), qToFontTQ2, "LEFT", 1, cProvider, prdb.questsCellPadding, 0, prdb.maxWidth, prdb.maxWidth, nil, "quest", uid, nil, tipAnchor)
				tip:SetCell(tempLine, 2, core.quests[uid].pqStatus, qToFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, prdb.statusMaxWidth, prdb.statusMaxWidth, nil, "quest", uid, nil, tipAnchor)
				if core.questsObjectives[uid] then
					for objective, qoData in pairs(core.questsObjectives[uid]) do
						if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and qoData.partyComplete == 0) or (prdb.hideCompletedObjective == 1 and qoData.playerComplete == 0) then
							tempLine = tip:AddLine()
							tip:SetCell(tempLine, 1, format("|cff%s%s|r", prdb.colorizeQuestObjective and qoData.qoColor or prdb.defaultQuestObjectiveColor, qoData.desc), qoToFontTQ2, "LEFT", 1, cProvider, prdb.objectivesCellPadding, 0, prdb.maxWidth, prdb.maxWidth, nil, "quest", uid, objective, tipAnchor)
							tip:SetCell(tempLine, 2, prdb.colorizeQuestObjectiveCounter and qoData.qoStatus or qoData.qoStatusNC, qoToFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, prdb.statusMaxWidth, prdb.statusMaxWidth, nil, "quest", uid, objective, tipAnchor)
						end
					end
				else
					tempLine = tip:AddLine()
					tip:SetCell(tempLine, 1, core.quests[uid].objective, qoToFontTQ2, "LEFT", 2, cProvider, prdb.objectivesCellPadding, prdb.rightCellPadding, prdb.maxWidth + prdb.statusMaxWidth, prdb.maxWidth + prdb.statusMaxWidth, nil, "quest", uid, nil, tipAnchor)
				end
			end
		end
	end
	tip:AddLine(" ")
	if prdb.showHints then
		for k, v in pairs(interactivityAction) do
			local action = prdb[k]
			if action ~= "nothing" then
				tempLine = tip:AddLine()
				tip:SetCell(tempLine, 1, format("|cffeda55f%s|r |cff19ff19%s.|r", v, interactivityList[action]), qoToFontTQ2, "LEFT", 2, cProvider, 0, prdb.rightCellPadding, prdb.maxWidth + prdb.statusMaxWidth, prdb.maxWidth + prdb.statusMaxWidth, nil, "hint")
			end
		end
	end
	tip:SmartAnchorTo(anchor)
	tip:UpdateScrolling(prdb.hasMaxHeight and prdb.maxHeight or nil)
	tip:Show()
end

function dataobj.OnEnter(self)
	if not tooltipIsShown then
		anchor = self
		tipAnchor = self:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
		module:updateTooltip(self)
	end
end

function dataobj.OnLeave(self)
	if tip and MouseIsOver(tip) then 
		tip:SetScript("OnUpdate", hideTooltip)
	else
		hideTooltip()
	end
end

local ldbOnClick = {
	openOptions = function()
		InterfaceOptionsFrame_OpenToCategory(core.optionsFrames.Profiles)
		InterfaceOptionsFrame_OpenToCategory(core.optionsFrames.tomQuest2)
	end,
	toggleQuestTracker = function()
		tQ2toggleQuestTracker()
	end,
	toggleAchievementTracker = function()
		tQ2toggleAchievementTracker()
	end,
	toggleAnnouncementModule = function()
		core:toggleAnnounce()
	end,
	toggleQuestsLog = function()
		ToggleFrame(QuestLogFrame)
	end,
	toggleSyncPartyModule = function()
		core:toggleSyncParty()
	end,
	toggleQuestsTrackerModule = function()
		core:toggleQuestsTracker()
	end,
	toggleAchievementTrackerModule = function()
		core:toggleAchievementTracker()
	end,
	nothing = function(cValue1)
		return
	end,
}

function dataobj.OnClick(self, button)
	if button == "RightButton" then
		if IsShiftKeyDown() then
			ldbOnClick[prdb.ldbMouseRightShift]()
		elseif IsControlKeyDown() then
			ldbOnClick[prdb.ldbMouseRightControl]()
		elseif IsAltKeyDown() then
			ldbOnClick[prdb.ldbMouseRightAlt]()
		else
			ldbOnClick[prdb.ldbMouseRight]()
		end
	elseif button == "LeftButton" then
		if IsShiftKeyDown() then
			ldbOnClick[prdb.ldbMouseLeftShift]()
		elseif IsControlKeyDown() then
			ldbOnClick[prdb.ldbMouseLeftControl]()
		elseif IsAltKeyDown() then
			ldbOnClick[prdb.ldbMouseLeftAlt]()
		else
			ldbOnClick[prdb.ldbMouseLeft]()
		end
	elseif button == "MiddleButton" then
		if IsShiftKeyDown() then
			ldbOnClick[prdb.ldbMouseMiddleShift]()
		elseif IsControlKeyDown() then
			ldbOnClick[prdb.ldbMouseMiddleControl]()
		elseif IsAltKeyDown() then
			ldbOnClick[prdb.ldbMouseMiddleAlt]()
		else
			ldbOnClick[prdb.ldbMouseMiddle]()
		end
	end
end