-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("questsTracker", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0")
local achievementTracker = core:GetModule("achievementTracker")

-----------------------
--- Library loading ---
-----------------------
local LQTr = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-QuestsTracker")
local LSM = LibStub("LibSharedMedia-3.0")
local QTIP = LibStub("LibQTip-1.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs, unpack, format, type, sub, tonumber = pairs, unpack, string.format, type, string.sub, tonumber
local ToggleFrame = ToggleFrame
local SetAbandonQuest, GetAbandonQuestItems, GetAbandonQuestName = SetAbandonQuest, GetAbandonQuestItems, GetAbandonQuestName
local SelectQuestLogEntry = SelectQuestLogEntry
local QuestLogPushQuest = QuestLogPushQuest
local MouseIsOver = MouseIsOver
local ItemRefTooltip = ItemRefTooltip
local GetQuestLogSelection = GetQuestLogSelection
local GetQuestLink = GetQuestLink
local CreateFrame, CreateFont, SetFont = CreateFrame, CreateFont, SetFont
local UnitIsDead = UnitIsDead
local GetSubZoneText, GetZoneText = GetSubZoneText, GetZoneText
local IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown = IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown
local ChatEdit_InsertLink = ChatEdit_InsertLink
local ShowUIPanel, HideUIPanel = ShowUIPanel, HideUIPanel

-----------------
--- Variables ---
-----------------
module.modName = LQTr["Quests Tracker"]
local prdb, chdb, options
local subZone, currentZone, oldZone = "", "", ""
local data = {true, true, true}
local autoUntrack, delayedHide = {}, {}
local anchorPosition = {[1] = "TOPLEFT", [2] = "TOP", [3] = "TOPRIGHT", [4] = "BOTTOMLEFT", [5] = "BOTTOM", [6] = "BOTTOMRIGHT"}
local autoTrack = {[1] = LQTr["Always"], [2] = LQTr["Below Max Level"], [3] = LQTr["Never"]}
local hideQuests = {[1] = LQTr["When you've completed the quest"], [2] = LQTr["When everyone in the party has completed the quest"], [3] = LQTr["Never"]}
local hideObjectives = {[1] = LQTr["When you've completed the objective"], [2] = LQTr["When everyone in the party has completed the objective"], [3] = LQTr["Never"]}
local outlines = {[""] = LQTr["None"], ["OUTLINE"] = LQTr["Outline"], ["THICKOUTLINE"] = LQTr["Thick Outline"]}
local defaultColors = {"defaultHeaderNameColor", "defaultHeaderLevelColor", "defaultQuestNameColor", "defaultQuestLevelColor", "defaultQuestObjectiveColor"}
local trackerAnchor, tracker = nil, nil
local hTrFontTQ2 = CreateFont("hTrFontTQ2")
local qTrFontTQ2 = CreateFont("qTrFontTQ2")
local qoTrFontTQ2 = CreateFont("qoTrFontTQ2")
local hideInCombat = nil
local interactivityList = {
	["chatLink"] = LQTr["Insert quest link in chat"],
	["objectiveProgress"] = LQTr["Insert quest objectives progress in chat"],
	["macroUpdate"] = LQTr["Update special quest item macro"],
	["questTooltip"] = LQTr["Display of quest tooltip"],
	["shareQuest"] = LQTr["Share quest"],
	["trackQuest"] = LQTr["Track quest"],
	["abandonQuest"] = LQTr["Abandon quest"],
	["displayQuest"] = LQTr["Diplay quest log"],
	["displayQuestInfo"] = LQTr["Display quest on map (external module needed)"],
	["displayCQIStartNPC"] = LQTr["Display of start npc in Cartographer quest Info"],
	["displayCQIEndNPC"] = LQTr["Display of end npc in Cartographer Quest Info"],
	["displayCQIObjective"] = LQTr["Display of objective in Cartographer Quest Info"],
	["displayMMObjective"] = LQTr["Display of objective in your map with MobMap"],
	["nothing"] = LQTr["Do nothing"],
}

----------------------------
--- Tracker Anchor Frame ---
----------------------------
local anchorSetPoint = {
	[1] = {point = "TOPLEFT", relPoint = "BOTTOMLEFT", x = 0, y = 24},
	[2] = {point = "TOP", relPoint = "BOTTOM", x = 0, y = 24},
	[3] = {point = "TOPRIGHT", relPoint = "BOTTOMRIGHT", x = 0, y = 24},
	[4] = {point = "BOTTOMLEFT", relPoint = "TOPLEFT", x = 0, y = -24},
	[5] = {point = "BOTTOM", relPoint = "TOP", x = 0, y = -24},
	[6] = {point = "BOTTOMRIGHT", relPoint = "TOPRIGHT", x = 0, y = -24}}

local anchor = CreateFrame("Button", tomQuest2QuestTrackerAnchor, UIParent)
anchor:SetHeight(24)
anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
local text = anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
text:SetPoint("CENTER")
text:SetText(LQTr["Tracker"])
anchor:SetWidth(text:GetStringWidth() + 8)
anchor:SetMovable(true)
anchor:RegisterForDrag("LeftButton")
anchor:SetClampedToScreen(true)
anchor:SetScript("OnDragStart", function(self)
	self:StartMoving()
end)
anchor:SetScript("OnDragStop", function(self)
	self:StopMovingOrSizing()
	prdb.point, prdb.x, prdb.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	module:updateQuestsTracker()
end)

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		point = "CENTER",
		x = 0,
		y = 0,
		onLogIn = false,
		onQuestAccepted = 3,
		onObjectiveProgress = true,
		unTrackInactive = false,
		unTrackDelay = 300,
		onSubZone = true,
		onZone = true,
		onDaily = true,
		onClass = true,
		showAnchor = true,
		anchorPosition = 2,
		toggleTracker = true,
		hideTrackerInCombat = true,
		hideCompletedQuest = 2,
		delayToHideCompletedQuest = 0,
		hideCompletedObjective = 2,
		onlyCurrentZone = false,
		limitQuestLevel = false,
		headerFont = LSM.DefaultMedia.font,
		headerFontSize = 12,
		headerFontOutline = "",
		questFont = LSM.DefaultMedia.font,
		questFontSize = 10,
		questFontOutline = "",
		questObjectiveFont = LSM.DefaultMedia.font,
		questObjectiveFontSize = 8,
		questObjectiveFontOutline = "",
		maxWidth = 300,
		statusMaxWidth = 100,
		hasMaxHeight = true,
		maxHeight = 600,
		scale = 1,
		showQuestItemIcon = true,
		questItemIconSize = 24,
		showItemsStartingQuest = true,
		showUsableQuestItemsNotAttached = true,
		questsCellPadding = 5,
		objectivesCellPadding = 10,
		rightCellPadding = 5,
		backDropColor = {0.09, 0.09, 0.09, 1},
		borderColor = {1, 1, 1, 1},
		colorizeQuestName = true,
		defaultQuestNameColor = "ffffff",
		colorizeQuestLevel = true,
		defaultQuestLevelColor = "ffffff",
		colorizeQuestObjective = true,
		defaultQuestObjectiveColor = "ffffff",
		colorizeQuestObjectiveCounter = true,
		displayHeader = true,
		autoCollapseHeader = true,
		collapseButtonSize = 24,
		colorizeHeaderName = true,
		defaultHeaderNameColor = "ffffff",
		hideZoneLevelRange = false,
		colorizeHeaderLevel = true,
		defaultHeaderLevelColor = "ffffff",
		disableInteraction = false,
		overTooltip = true,
		mouseRightShift = "chatLink",
		mouseRightControl = "objectiveProgress",
		mouseRightAlt = "macroUpdate",
		mouseRight = "questTooltip",
		mouseLeftShift = "trackQuest",
		mouseLeftControl = "shareQuest",
		mouseLeftAlt = "abandonQuest",
		mouseLeft = "displayQuest",
		mouseMiddleShift = "nothing",
		mouseMiddleControl = "nothing",
		mouseMiddleAlt = "nothing",
		mouseMiddle = "nothing",
	},
}

---------------
--- Options ---
---------------
local optGetter, optSetter, colorGetter, colorSetter, hexColorGetter, hexColorSetter, anchorSetter, combatSetter, maxLDGetter, maxLDSetter, headFontSetter, questFontSetter, qoFontSetter, interactionDisabled
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		QTIP:Release(tracker)
		tracker = nil
		module:updateQuestsTracker()
	end
	function colorGetter(info)
		return unpack(prdb[info[#info]])
	end
	function colorSetter(info, arg1, arg2, arg3, arg4)
		prdb[info[#info]][1] = arg1
		prdb[info[#info]][2] = arg2
		prdb[info[#info]][3] = arg3
		prdb[info[#info]][4] = arg4
		module:updateQuestsTracker()
	end
	function hexColorGetter(info)
		return tonumber(sub(prdb[info[#info]], 1, 2), 16) / 255, tonumber(sub(prdb[info[#info]], 3, 4), 16) / 255, tonumber(sub(prdb[info[#info]], 5, 6), 16) / 255
	end
	function hexColorSetter(info, arg1, arg2, arg3)
		prdb[info[#info]] = format("%02x%02x%02x", arg1*255, arg2*255, arg3*255)
		core:computeQuestsStatus()
	end
	function anchorSetter(info, value)
		prdb[info[#info]] = value
		if value then
			anchor:Show()
		else
			anchor:Hide()
		end
		module:updateQuestsTracker()
	end
	function combatSetter(info, value)
		prdb[info[#info]] = value
		hideInCombat = nil
		if value then
			core.RegisterCallback(module, "outOfCombat")
			core.RegisterCallback(module, "inCombat")
		else
			core.UnregisterCallback(module, "outOfCombat")
			core.UnregisterCallback(module, "inCombat")
		end
	end
	function maxLDGetter(info)
		return core.db.profile.quests[info[#info]]
	end
	function maxLDSetter(info, value)
		core.db.profile.quests[info[#info]] = value
		module:questsWeightedLevelUpdated()
		module:updateQuestsTracker()
	end
	function headFontSetter(info, value)
		prdb[info[#info]] = value
		hTrFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
		module:updateQuestsTracker()
	end
	function questFontSetter(info, value)
		prdb[info[#info]] = value
		qTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
		module:updateQuestsTracker()
	end
	function qoFontSetter(info, value)
		prdb[info[#info]] = value
		qoTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
		module:updateQuestsTracker()
	end
	function interactionDisabled()
		if prdb.disableInteraction then
			return true
		else
			return false
		end
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			childGroups = "tab",
			name = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				autotrack = {
					type = "group",
					name = LQTr["Auto track"],
					args = {
						onLogIn = {
							name = LQTr["Track all quests on log in"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						onQuestAccepted = {
							name = LQTr["Auto track quests when you accept them"],
							type = "select",
							values = autoTrack,
							width = "full",
							order = 2,
						},
						onObjectiveProgress = {
							name = LQTr["Auto track quests on objective progress"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						unTrackInactive = {
							name = LQTr["Auto untrack inactive quests that were tracked on objective progress"],
							type = "toggle",
							width = "full",
							order = 4,
						},
						unTrackDelay = {
							name = LQTr["Auto untrack delay (in minutes)"],
							type = "range",
							min = 1,
							max = 30,
							step = 0.5,
							width = "full",
							order = 5,
						},
						onSubZone = {
							name = LQTr["Auto track quests on sub zone change"],
							type = "toggle",
							width = "full",
							order = 6,
						},
						onZone = {
							name = LQTr["Auto track quests on zone change"],
							type = "toggle",
							width = "full",
							order = 7,
						},
						onDaily = {
							name = LQTr["Auto track daily quests"],
							type = "toggle",
							width = "full",
							order = 8,
						},
						onClass = {
							name = LQTr["Auto track class quests"],
							type = "toggle",
							width = "full",
							order = 9,
						},
					},
				},
				display = {
					type = "group",
					name = LQTr["Display"],
					args = {
						showAnchor = {
							name = LQTr["Show quest tracker anchor"],
							type = "toggle",
							width = "full",
							set = anchorSetter,
							order = 1,
						},
						anchorPosition = {
							name = LQTr["Anchor position"],
							type = "select",
							values = anchorPosition,
							width = "full",
							order = 2,
						},
						displayHeader = {
							name = LQTr["Display quests by zone"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						autoCollapseHeader = {
							name = LQTr["Automatically expand /collapse zone header on zone change"],
							type = "toggle",
							width = "full",
							order = 4,
						},
						toggleTracker = {
							name = LQTr["Display tracker"],
							type = "toggle",
							width = "full",
							order = 5,
						},
						hideTrackerInCombat = {
							name = LQTr["Hide quest tracker in combat"],
							type = "toggle",
							set = combatSetter,
							width = "full",
							order = 6,
						},
						hideCompletedQuest = {
							name = LQTr["Hide completed quest"],
							type = "select",
							values = hideQuests,
							width = "full",
							order = 7,
						},
						delayToHideCompletedQuest = {
							name = LQTr["Delay to hide completed quest"],
							type = "range",
							min = 0,
							max = 60,
							step = 0.5,
							width = "full",
							order = 8,
						},
						hideCompletedObjective = {
							name = LQTr["Hide completed quests Objectives"],
							type = "select",
							values = hideObjectives,
							width = "full",
							order = 9,
						},
						onlyCurrentZone = {
							name = LQTr["Hide quests that are not in the current zone"],
							type = "toggle",
							width = "full",
							order = 10,
						},
						limitQuestLevel = {
							name = LQTr["Hide high level quests (You'll have to choose the maximum level difference)"],
							type = "toggle",
							width = "full",
							order = 11,
						},
						maxLevelDiff = {
							name = LQTr["Maximum level difference between the weighted quest level and your level"],
							type = "range",
							min = 0,
							max = 10,
							step = 1,
							get = maxLDGetter,
							set = maxLDSetter,
							width = "full",
							order = 12,
						},
						hideZoneLevelRange = {
							name = LQTr["Hide zone level range"],
							type = "toggle",
							width = "full",
							order = 13,
						},
					},
				},
				fonts = {
					type = "group",
					name = LQTr["Tracker fonts"],
					args = {
						zone = {
							type = "group",
							name = LQTr["Zone"],
							set = headFontSetter,
							inline = true,
							order = 1,
							args = {
								headerFont = {
									name = LQTr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								headerFontSize = {
									name = LQTr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								headerFontOutline = {
									name = LQTr["Font outline"],
									type = "select",
									values = outlines,
									width = "full",
									order = 3,
								},
							},
						}, 
						quests = {
							type = "group",
							name = LQTr["Quests"],
							set = questFontSetter,
							inline = true,
							order = 2,
							args = {
								questFont = {
									name = LQTr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								questFontSize = {
									name = LQTr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								questFontOutline = {
									name = LQTr["Font outline"],
									type = "select",
									values = outlines,
									width = "full",
									order = 3,
								},
							},
						},
						questsObjectives = {
							type = "group",
							name = LQTr["Quests objectives"],
							set = qoFontSetter,
							inline = true,
							order = 3,
							args = {
								questObjectiveFont = {
									name = LQTr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								questObjectiveFontSize = { 
									name = LQTr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								questObjectiveFontOutline = {
									name = LQTr["Font outline"],
									type = "select",
									values = outlines,
									width = "full",
									order = 3,
								},
							},
						},
					},
				},
				appearance = {
					type = "group",
					name = LQTr["Tracker appearance"],
					args = {
						scale = {
							name = LQTr["Quests Tracker scale"],
							type = "range",
							min = 0,
							max = 1,
							step = 0.1,
							width = "full",
							order = 1,
						},
						backDropColor = {
							name = LQTr["Quests Tracker background color"],
							type = "color",
							hasAlpha = true,
							get = colorGetter,
							set = colorSetter,
							width = "full",
							order = 2,
						},
						borderColor = {
							name = LQTr["Quests Tracker border color"],
							type = "color",
							hasAlpha = true,
							get = colorGetter,
							set = colorSetter,
							width = "full",
							order = 3,
						},
						hasMaxHeight = {
							name = LQTr["Set quests tracker maximum height after which it will be scrollable"],
							type = "toggle",
							width = "full",
							order = 4,
						},
						maxHeight = {
							name = LQTr["Quests tracker max height"],
							type = "range",
							min = 100,
							max = 2000,
							step = 5,
							width = "full",
							order = 5,
						},
						collapseButtonSize = {
							name = LQTr["Collapse Button Size"],
							type = "range",
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 6,
						},
						questsCellPadding = {
							name = LQTr["Quests cell padding"],
							type = "range",
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 7,
						},
						objectivesCellPadding = {
							name = LQTr["Quests objectives cell padding"],
							type = "range",
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 8,
						},
						rightCellPadding = {
							name = LQTr["Right-most cell right padding to avoid cell clipping"],
							type = "range",
							min = 0,
							max = 10,
							step = 1,
							width = "full",
							order = 9,
						},
						maxWidth = {
							name = LQTr["Quests Tracker column max width"],
							type = "range",
							min = 50,
							max = 600,
							step = 5,
							width = "full",
							order = 10,
						},
						statusMaxWidth = {
							name = LQTr["Quests Tracker status column max width"],
							type="range",
							min = 30,
							max = 200,
							step = 5,
							width = "full",
							order = 11,
						},
					},
				},
				color = {
					type = "group",
					name = LQTr["Colors"],
					args = {
						colorizeHeaderName = {
							name = LQTr["Colorize Quests Header Name"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						defaultHeaderNameColor = {
							name = LQTr["Default header name color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 2,
						},
						colorizeHeaderLevel = {
							name = LQTr["Colorize Quests Header Level"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						defaultHeaderLevelColor = {
							name = LQTr["Default header level color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 4,
						},
						colorizeQuestName = {
							name = LQTr["Colorize Quests Name"],
							type = "toggle",
							width = "full",
							order = 5,
						},
						defaultQuestNameColor = {
							name = LQTr["Default quest name color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 6,
						},
						colorizeQuestLevel = {
							name = LQTr["Colorize Quests Level"],
							type = "toggle",
							width = "full",
							order = 7,
						},
						defaultQuestLevelColor = {
							name = LQTr["Default quest level color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 8,
						},
						colorizeQuestObjective = {
							name = LQTr["Colorize Quests Objectives"],
							type = "toggle",
							width = "full",
							order = 9,
						},
						defaultQuestObjectiveColor = {
							name = LQTr["Default quest objective color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 10,
						},
						colorizeQuestObjectiveCounter = {
							name = LQTr["Colorize quests objectives counter"],
							type = "toggle",
							width = "full",
							order = 11,
						},
					},
				},
				questsItems = {
					type = "group",
					name = LQTr["Quests items"],
					args = {
						showQuestItemIcon = {
							name = LQTr["Show special quest item icon behind quest name"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						questItemIconSize = {
							name = LQTr["Quest item icon size"],
							type = "range",
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 2,
						},
						showItemsStartingQuest = {
							name = LQTr["Show items starting a quest in the tracker"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						showUsableQuestItemsNotAttached = {
							name = LQTr["Show usable quest item not linked to a quest"],
							type = "toggle",
							width = "full",
							order = 4,
						},
					},
				},
				trackerInteraction = {
					type = "group",
					name = LQTr["Tracker interaction"],
					args = {
						disableInteraction = {
							type = "toggle",
							name = LQTr["Completely disable tracker interaction, should not be used when display by category is used or if you wish to make the it scrollable."],
							width = "full",
								order = 1,
						},
						overTooltip = {
							name = LQTr["Display quest tooltip on mouse over"],
							type = "toggle",
							width = "full",
							hidden = interactionDisabled,
							order = 2,
						},
						mouseRightShift = {
							name = LQTr["Action on right shift click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 3,
						},
						mouseRightControl = {
							name = LQTr["Action on right control click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 4,
						},
						mouseRightAlt = {
							name = LQTr["Action on right alt click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 5,
						},
						mouseRight = {
							name = LQTr["Action on right click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 6,
						},
						mouseLeftShift = {
							name = LQTr["Action on left shift click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 7,
						},
						mouseLeftControl = {
							name = LQTr["Action on left control click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 8,
						},
						mouseLeftAlt = {
							name = LQTr["Action on left alt click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 9,
						},
						mouseLeft = {
							name = LQTr["Action on left click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 10,
						},
						mouseMiddleShift = {
							name = LQTr["Action on middle shift click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 11,
						},
						mouseMiddleControl = {
							name = LQTr["Action on middle control click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 12,
						},
						mouseMiddleAlt = {
							name = LQTr["Action on middle alt click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 13,
						},
						mouseMiddle = {
							name = LQTr["Action on middle click"],
							type = "select",
							values = interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 14,
						},
					},
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("questsTracker", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	self:SetEnabledState(core:getModuleEnabled("questsTracker"))
	core:registerModuleOptions("questsTracker", getOptions, module.modName)
	chdb.questsTracked = chdb.questsTracked or {}
	chdb.questsSubZoneAutoTracked = chdb.questsSubZoneAutoTracked or {}
	chdb.questsZoneAutoTracked = chdb.questsZoneAutoTracked or {}
	chdb.zoneDisplay = chdb.zoneDisplay or {}
	core.RegisterCallback(self, "updateDBtoV3")
	LSM.RegisterCallback(self, "LibSharedMedia_Registered", "updateUsedMedia")
end

function module:updateDBtoV3(event)
	for k in pairs(defaultColors) do
		if type(prdb[k]) == "table" then
			prdb[k] = format("%02x%02x%02x", (prdb[k][1] or 0)*255, (prdb[k][2] or 0)*255, (prdb[k][3] or 0)*255)
		end
	end
end

function module:OnEnable()
	hTrFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	qTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	qoTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
	subZone = GetSubZoneText() or ""
	currentZone = GetZoneText() or ""
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if prdb.hideTrackerInCombat then
		core.RegisterCallback(self, "outOfCombat")
		core.RegisterCallback(self, "inCombat")
	end
	self:RegisterEvent("ZONE_CHANGED", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED_INDOORS", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "zoneChanged")
	self:SecureHook("AddQuestWatch")
	self:SecureHook("RemoveQuestWatch")
	self:SecureHook("GetNumQuestWatches")
	core.RegisterCallback(self, "initQuestsTracker")
	core.RegisterCallback(self, "endPlayerQuest")
	core.RegisterCallback(self, "endPartyQuest")
	core.RegisterCallback(self, "playerQuestComplete", "playerQuestComplete")
	core.RegisterCallback(self, "playerQuestUnComplete", "playerQuestUnComplete")
	core.RegisterCallback(self, "playerQuestFailed", "questFailed")
	core.RegisterCallback(self, "partyQuestComplete", "partyQuestComplete")
	core.RegisterCallback(self, "partyQuestUnComplete", "partyQuestUnComplete")
	core.RegisterCallback(self, "partyQuestFailed", "questFailed")
	core.RegisterCallback(self, "playerQuestObjectiveProgress", "questObjectiveProgress")
	core.RegisterCallback(self, "partyQuestObjectiveProgress", "questObjectiveProgress")
	core.RegisterCallback(self, "playerQuestLost", "questLost")
	core.RegisterCallback(self, "partyQuestLost", "questLost")
	core.RegisterCallback(self, "toggleTrackedQuest")
	core.RegisterCallback(self, "questObjectiveAdded")
	core.RegisterCallback(self, "questDescriptionAdded")
	core.RegisterCallback(self, "updateQuestsTracker")
end

function module:OnDisable()
	anchor:Hide()
	QTIP:Release(tracker)
	tracker = nil
	core.UnregisterAllCallbacks(self)
	self:Refresh()
	core.cbh:Fire("updateAchievementTracker")
end

function module:Refresh()
	prdb = self.db.profile
	chdb = self.db.char
	if not self:IsEnabled() then return end
	chdb.questsTracked = chdb.questsTracked or {}
	chdb.questsSubZoneAutoTracked = chdb.questsSubZoneAutoTracked or {}
	chdb.questsZoneAutoTracked = chdb.questsZoneAutoTracked or {}
	chdb.zoneDisplay = chdb.zoneDisplay or {}
	hTrFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	qTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	qoTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if prdb.hideTrackerInCombat then
		core.RegisterCallback(self, "outOfCombat")
		core.RegisterCallback(self, "inCombat")
	end
end

function module:updateUsedMedia(event, mediatype, key)
	if mediatype == "font" then
		local updateQuestTracker
		if key == prdb.headerFont then
			updateQuestTracker = true
			hTrFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
		end
		if key == prdb.questFont then
			updateQuestTracker = true
			qTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
		end
		if key == prdb.questObjectiveFont then
			updateQuestTracker = true
			qoTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
		end
		if updateQuestTracker then
			module:updateQuestsTracker()
		end
	end
end

-------------------
--- zoneChanged ---
-------------------
function module:zoneChanged()
	if UnitIsDead("player") then return end
	local trackerNeedUpdate = false
	subZone = GetSubZoneText() or ""
	currentZone = GetZoneText() or ""
	if prdb.onZone and currentZone ~= oldZone then
		for uid in pairs(chdb.questsZoneAutoTracked) do
			if core.quests[uid] then
				trackerNeedUpdate = true
				RemoveQuestWatch(core.quests[uid].id, uid, true, true)
			else
				chdb.questsZoneAutoTracked[uid] = nil
			end
		end
		for uid, qData in pairs(core.quests) do
			if not chdb.questsTracked[uid] and qData.zone == currentZone then
				chdb.questsZoneAutoTracked[uid] = true
				trackerNeedUpdate = true
				AddQuestWatch(core.quests[uid].id, nil, uid, true)
			end
		end
	end
	if prdb.onSubZone and subZone ~= "" then
		for uid in pairs(chdb.questsSubZoneAutoTracked) do
			if core.quests[uid] then
				trackerNeedUpdate = true
				RemoveQuestWatch(core.quests[uid].id, uid, true, true)
			else
				chdb.questsSubZoneAutoTracked[uid] = nil
			end
		end
		for uid, qData in pairs(core.quests) do
			if not chdb.questsTracked[uid] and ((qData.objective and qData.objective:find(subZone, 1, true)) or (qData.description and qData.description:find(subZone, 1, true))) then
				chdb.questsSubZoneAutoTracked[uid] = true
				trackerNeedUpdate = true
				AddQuestWatch(core.quests[uid].id, nil, uid, true)
			end
		end
	end
	if prdb.displayHeader and prdb.autoCollapseHeader then
		for zone, zData in pairs(core.zones) do
			if currentZone == zone then
				chdb.zoneDisplay[zone] = 1
			elseif zData.zone then
				chdb.zoneDisplay[zone] = 0
			end
		end
		trackerNeedUpdate = true
	end
	if trackerNeedUpdate then
		module:updateQuestsTracker()
	end
	for uid in pairs(autoUntrack) do
		if not chdb.questsTracked[uid] then
			module:CancelTimer(autoUntrack[uid], true)
			autoUntrack[uid] = nil
		end
	end
	oldZone = currentZone
end

-------------
--- Hooks ---
-------------
function module:AddQuestWatch(id, timeToTrack, uid, batch)
	if not uid then
		uid = core.questsIdToUid[id]
		if not uid then return end
	end
	chdb.questsTracked[uid] = true
	core.quests[uid].tracked = true
	if not batch then
		module:updateQuestsTracker()
	end
	if QuestLogFrame:IsVisible() then
		QuestLog_Update()
	end
end

function module:unTrackDelay(uid)
	if not uid or not core.quests[uid] or not core.quests[uid].id then return end
	RemoveQuestWatch(core.quests[uid].id, uid)
	autoUntrack[uid] = nil
end

function module:RemoveQuestWatch(id, uid, batch, leaveTimer)
	if not uid then
		uid = core.questsIdToUid[id]
		if not uid then return end
	end
	if chdb.questsTracked[uid] then
		chdb.questsTracked[uid] = nil
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		core.quests[uid].tracked = nil
		if not batch then
			module:updateQuestsTracker()
		end
	end
	if autoUntrack[uid] and not leaveTimer then
		module:CancelTimer(autoUntrack[uid], true)
		autoUntrack[uid] = nil
	end
	if QuestLogFrame:IsVisible() then
		QuestLog_Update()
	end
end

function module:GetNumQuestWatches()
	local numTQ = 0
	for uid in pairs(chdb.questsTracked) do
		if core.quests[uid] then
			numTQ = numTQ + 1
		end
	end
	return numTQ
end

function module:isQuestTracked(uid)
	if chdb.questsTracked[uid] then
		return true
	else
		return false
	end
end

-----------------
--- Callbacks ---
-----------------
function module:initQuestsTracker(event)
	for uid in pairs(chdb.questsTracked) do
		if not core.quests[uid] then
			chdb.questsTracked[uid] = nil
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
		else
			AddQuestWatch(core.quests[uid].id, nil, uid, true)
		end
	end
	module:updateQuestsTracker()
end

function module:endPlayerQuest(event, first, commHandler, uid, started)
	if first then
		if not chdb.questsTracked[uid] and prdb.onLogIn then
			AddQuestWatch(core.quests[uid].id, nil, uid, true)
		end
	elseif prdb.onQuestAccepted == 1 or (prdb.onQuestAccepted == 2 and core.playerLevel ~= MAX_PLAYER_LEVEL) or (prdb.onObjectiveProgress and started == 1) or (prdb.onDaily and core.quests[uid].daily == 1) or (prdb.onClass and core.quests[uid].class) then
		if not chdb.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid)
		else
			if autoUntrack[uid] and (prdb.onQuestAccepted == 1 or (prdb.onQuestAccepted == 2 and core.playerLevel ~= MAX_PLAYER_LEVEL) or (prdb.onDaily and core.quests[uid].daily == 1) or (prdb.onClass and core.quests[uid].class)) then
				module:CancelTimer(autoUntrack[uid], true)
				autoUntrack[uid] = nil
			end
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
			module:updateQuestsTracker()
		end
	elseif prdb.onZone and core.quests[uid].zone == currentZone then
		if not chdb.questsTracked[uid] then
			chdb.questsZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil, true)
		else
			chdb.questsSubZoneAutoTracked[uid] = nil
			module:updateQuestsTracker()
		end
	elseif prdb.onSubZone and not chdb.questsTracked[uid] and subZone ~= "" and ((core.quests[uid].objective and core.quests[uid].objective:find(subZone, 1, true)) or (core.quests[uid].description and core.quests[uid].description:find(subZone, 1, true))) then
		chdb.questsSubZoneAutoTracked[uid] = true
		AddQuestWatch(core.quests[uid].id, nil, uid, nil, true)
	elseif chdb.questsTracked[uid] then
		module:updateQuestsTracker()
	end
end

function module:endPartyQuest(event, sender, uid, started, batch)
	if batch then
		if not chdb.questsTracked[uid] and prdb.onLogIn then
			AddQuestWatch(core.quests[uid].id, nil, uid, true)
		end
	elseif prdb.onQuestAccepted == 1 or (prdb.onQuestAccepted == 2 and core.playerLevel ~= MAX_PLAYER_LEVEL) or (prdb.onObjectiveProgress and started == 1) or (prdb.onDaily and core.quests[uid].daily == 1) then
		if chdb.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid)
		else
			if autoUntrack[uid] and (prdb.onQuestAccepted == 1 or (prdb.onQuestAccepted == 2 and core.playerLevel ~= MAX_PLAYER_LEVEL) or (prdb.onDaily and core.quests[uid].daily == 1)) then
				module:CancelTimer(autoUntrack[uid], true)
				autoUntrack[uid] = nil
			end
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
			module:updateQuestsTracker()		
		end
	elseif chdb.questsTracked[uid] then
		module:updateQuestsTracker()
	end
end

function module:playerQuestComplete(event, first, commHandler, uid, sender)
	local completeTime = GetTime()
	if autoUntrack[uid] then
		module:CancelTimer(autoUntrack[uid], true)
		autoUntrack[uid] = nil
	end
	core.quests[uid].playerCompleteTime = completeTime
	if core.quests[uid].partyComplete == 1 then
		core.quests[uid].partyCompleteTime = completeTime
	end
	if chdb.questsTracked[uid] and prdb.hideCompleteQuests ~= 3 and prdb.delayToHideCompletedQuest > 0 then
		if delayedHide[uid] then
			module:CancelTimer(delayedHide[uid], true)
			delayedHide[uid] = nil
		end
		delayedHide[uid] = module:ScheduleTimer("updateQuestsTracker", prdb.delayToHideCompletedQuest, uid)
	elseif chdb.questsTracked[uid] then
		module:updateQuestsTracker()
	end
end

function module:partyQuestComplete(event, first, commHandler, uid, sender)
	if core.quests[uid].partyComplete == 1 then
		if autoUntrack[uid] then
			module:CancelTimer(autoUntrack[uid], true)
			autoUntrack[uid] = nil
		end
		local completeTime = GetTime()
		core.quests[uid].partyCompleteTime = completeTime
		if chdb.questsTracked[uid] and prdb.hideCompleteQuests ~= 3 and prdb.delayToHideCompletedQuest > 0 then
			if delayedHide[uid] then
				module:CancelTimer(delayedHide[uid], true)
				delayedHide[uid] = nil
			end
			delayedHide[uid] = module:ScheduleTimer("updateQuestsTracker", prdb.delayToHideCompletedQuest, uid)
		elseif chdb.questsTracked[uid] then
			module:updateQuestsTracker()
		end
	elseif chdb.questsTracked[uid] then
		module:updateQuestsTracker()
	end
end

function module:playerQuestUnComplete(event, first, commHandler, uid, sender)
	core.quests[uid].playerCompleteTime = nil
	core.quests[uid].partyCompleteTime = nil
	module:updateQuestsTracker()
end

function module:partyQuestUnComplete(event, first, commHandler, uid, sender)
	core.quests[uid].partyCompleteTime = nil
	module:updateQuestsTracker()
end

function module:questFailed(event, first, commHandler, uid, sender)
	if chdb.questsTracked[uid] then
		if autoUntrack[uid] then
			module:CancelTimer(autoUntrack[uid], true)
			autoUntrack[uid] = nil
		end
		module:updateQuestsTracker()
	end
end

function module:questObjectiveProgress(event, first, commHandler, uid, objective, got, nGot, sender)
	if prdb.onObjectiveProgress then
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		if not chdb.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid)
			if prdb.unTrackInactive and prdb.unTrackDelay > 0 then
				if autoUntrack[uid] then
					module:CancelTimer(autoUntrack[uid], true)
					autoUntrack[uid] = nil
				end
				autoUntrack[uid] = module:ScheduleTimer("unTrackDelay", prdb.unTrackDelay * 60, uid)
			elseif prdb.unTrackInactive then
				RemoveQuestWatch(core.quests[uid].id, uid)
			end
		else
			if autoUntrack[uid] then
				module:CancelTimer(autoUntrack[uid], true)
				autoUntrack[uid] = nil
				if prdb.unTrackInactive and prdb.unTrackDelay > 0 then
					autoUntrack[uid] = module:ScheduleTimer("unTrackDelay", prdb.unTrackDelay * 60, uid)
				elseif prdb.unTrackInactive then
					RemoveQuestWatch(core.quests[uid].id, uid)
				end
			end
			module:updateQuestsTracker()
		end
	elseif chdb.questsTracked[uid] then
		module:updateQuestsTracker()
	end
end

function module:questLost(event, first, commHandler, uid, abandoning, title, sender)
	if not core.quests[uid] and chdb.questsTracked[uid] then
		if autoUntrack[uid] then
			module:CancelTimer(autoUntrack[uid], true)
			autoUntrack[uid] = nil
		end
		chdb.questsTracked[uid] = nil
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		module:updateQuestsTracker()
	elseif core.quests[uid] and chdb.questsTracked[uid] then
		module:updateQuestsTracker()
	end
end

function module:toggleTrackedQuest(event, uid)
	if chdb.questsTracked[uid] then
		RemoveQuestWatch(core.quests[uid].id, uid)
	elseif not chdb.questsTracked[uid] then
		AddQuestWatch(core.quests[uid].id, nil, uid)
	end
end

function module:questObjectiveAdded(event, uid)
	if prdb.onSubZone and subZone ~= "" then
		if not chdb.questsTracked[uid] and core.quests[uid].objective and core.quests[uid].objective:find(subZone, 1, true) then
			chdb.questsSubZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil, true)
		elseif chdb.questsTracked[uid] then
			module:updateQuestsTracker()
		end
	elseif chdb.questsTracked[uid] then
		module:updateQuestsTracker()
	end
end

function module:questDescriptionAdded(event, uid)
	if prdb.onSubZone and subZone ~= "" then
		if not chdb.questsTracked[uid] and core.quests[uid].description	and core.quests[uid].description:find(subZone, 1, true) then
			chdb.questsSubZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil, true)
		elseif chdb.questsTracked[uid] then
			module:updateQuestsTracker()
		end
	elseif chdb.questsTracked[uid] then
		module:updateQuestsTracker()
	end
end

function module:inCombat(event)
	hideInCombat = true
	if tracker then tracker:Hide() end
end

function module:outOfCombat(event)
	hideInCombat = nil
	module:updateQuestsTracker()
end

---------------------------------------------------
--- QTIP Mouse Handler and Custom Cell Provider ---
---------------------------------------------------
local interactivity = {
	chatLink = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			ChatFrameEditBox:Show()
			ChatEdit_InsertLink(GetQuestLink(core.quests[cValue1].id))
		end
	end,
	objectiveProgress = function(cValue1)
		if core.quests[cValue1].numObjectives > 0 and core.quests[cValue1].complete[core.playerName] then
			ChatFrameEditBox:Show()
			for objective, qoData in pairs(core.questsObjectives[cValue1]) do
				local got, _ = qoData.got[core.playerName]:match("^(.-)~(%d-)$")
				ChatFrameEditBox:Insert(format("{%s %s/%s} ", qoData.desc, got, qoData.need))
			end
		end
	end,
	macroUpdate = function(cValue1)
		core:generateSpecialItemMacro(cValue1)
	end,
	questTooltip = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			ItemRefTooltip:ClearLines()
			ShowUIPanel(ItemRefTooltip)
			if not ItemRefTooltip:IsVisible() then
				ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
			end
			local link = GetQuestLink(core.quests[cValue1].id)
			if link then
				ItemRefTooltip:SetHyperlink(link)
			end
		end
	end,
	shareQuest = function(cValue1)
		if core.numPartyMembers > 0 and core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id and core.quests[cValue1].sharable then
			local wasSelected = GetQuestLogSelection()
			SelectQuestLogEntry(core.quests[cValue1].id)
			QuestLogPushQuest()
			SelectQuestLogEntry(wasSelected)
		end
	end,
	trackQuest = function(cValue1)
		module:toggleTrackedQuest(nil, cValue1)
	end,
	abandonQuest = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			local wasSelected = GetQuestLogSelection()
			SelectQuestLogEntry(core.quests[cValue1].id)
			SetAbandonQuest()
			local items = GetAbandonQuestItems()
			if items then
				StaticPopup_Hide("ABANDON_QUEST")
				StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
			else
				StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
				StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
			end
			SelectQuestLogEntry(wasSelected)
		end
	end,
	displayQuest = function(cValue1)
		core:displayQuestInformations(cValue1)
	end,
	displayQuestInfo = function(cValue1, cValue2)
		if cValue2 then
			core.cbh:Fire("toggleQuestInfo", cValue1, cValue2)
		end
	end,
	displayCQIStartNPC = function(cValue1)
		if Cartographer_QuestInfo then
			local q = Cartographer_QuestInfo:GetQuest(cValue1)
			if q and q.start_npc then
				data.quest = q.title_full
				data.obj = nil
				data.npc = q.start_npc
				Cartographer_QuestInfo:OnButtonClick(nil, "start", "LeftButton", data)
			end
		end
	end,
	displayCQIEndNPC = function(cValue1)
		if Cartographer_QuestInfo then
			local q = Cartographer_QuestInfo:GetQuest(cValue1)
			if q and q.end_npc then
				data.quest = q.title_full
				data.obj = nil
				data.npc = q.end_npc
				Cartographer_QuestInfo:OnButtonClick(nil, "end", "LeftButton", data)
			end
		end
	end,
	displayCQIObjective = function(cValue1, cValue2)
		if cValue2 and Cartographer_QuestInfo then
			local q = Cartographer_QuestInfo:GetQuest(cValue1)
			if q and q.objs then
				for _, obj in pairs(q.objs) do
					if obj.title:find(core.questsObjectives[cValue1][cValue2].desc, 1, true) then
						data.quest = q.title_full
						data.obj = obj
						data.npc = nil
					end
				end
				if data.obj ~= nil then
					Cartographer_QuestInfo:OnButtonClick(nil, "obj", "LeftButton", data)
				end
			end
		end
	end,
	displayMMObjective = function(cValue1, cValue2)
		if cValue2 and MobMapDotParentFrame and MobMap_ParseQuestObjective then
			MobMap_ParseQuestObjective(core.questsObjectives[cValue1][cValue2].desc)
		end
	end,
	nothing = function(cValue1, cValue2)
		return
	end
}

local questsOnMouseDown = function(cell, button)
	local category, cType, cValue1, cValue2, anchor = cell.category, cell.cType, cell.cValue1, cell.cValue2, cell.anchor
	if category == "achievements" then
		achievementTracker:achievementOnMouseDown(cell, button)
	elseif category == "quests" then
		if cType == "zone" then
			if chdb.zoneDisplay[cValue1] == 1 then
				chdb.zoneDisplay[cValue1] = 0
			else
				chdb.zoneDisplay[cValue1] = 1
			end
			module:updateQuestsTracker()
		elseif cType == "quest" then
			if button == "RightButton" then
				if IsShiftKeyDown() then
					interactivity[prdb.mouseRightShift](cValue1, cValue2)
				elseif IsControlKeyDown() then
					interactivity[prdb.mouseRightControl](cValue1, cValue2)
				elseif IsAltKeyDown() then
					interactivity[prdb.mouseRightAlt](cValue1, cValue2)
				else
					interactivity[prdb.mouseRight](cValue1, cValue2)
				end
			elseif button == "LeftButton" then
				if IsShiftKeyDown() then
					interactivity[prdb.mouseLeftShift](cValue1, cValue2)
				elseif IsControlKeyDown() then
					interactivity[prdb.mouseLeftControl](cValue1, cValue2)
				elseif IsAltKeyDown() then
					interactivity[prdb.mouseLeftAlt](cValue1, cValue2)
				else
					interactivity[prdb.mouseLeft](cValue1, cValue2)
				end
			elseif button == "MiddleButton" then
				if IsShiftKeyDown() then
					interactivity[prdb.mouseMiddleShift](cValue1, cValue2)
				elseif IsControlKeyDown() then
					interactivity[prdb.mouseMiddleControl](cValue1, cValue2)
				elseif IsAltKeyDown() then
					interactivity[prdb.mouseMiddleAlt](cValue1, cValue2)
				else
					interactivity[prdb.mouseMiddle](cValue1, cValue2)
				end
			end
		elseif cType == "questItem" then
			core:generateSpecialItemMacro(nil, cValue1)
		end
	end
end
local questsOnEnter = function(cell)
	local category, cType, cValue1, cValue2, anchor = cell.category, cell.cType, cell.cValue1, cell.cValue2, cell.anchor
	if category == "achievements" then
		achievementTracker:achievementOnEnter(cell)
	elseif category == "quests" then
		if prdb.overTooltip and cType == "quest" then
			if cValue1 and cValue2 and core.questsObjectives[cValue1] and core.questsObjectives[cValue1][cValue2].cType == "object" and core.questsItems[core.questsObjectives[cValue1][cValue2].desc].mobDrop then
				GameTooltip:SetOwner(tracker, anchor)
				GameTooltip:AddLine(LQTr["|cffffffffDrop from:|r"])
				for mobName in pairs(core.questsItems[core.questsObjectives[cValue1][cValue2].desc].mobDrop) do
					local zones = ""
					for zone in pairs(core.questsMobsItems[mobName].mobLoc) do
						zones = zones == "" and zone or format("%s, %s", zones, zone)
					end
					if core.questsMobsItems[mobName].items[core.questsObjectives[cValue1][cValue2].desc].dropRate then
						GameTooltip:AddDoubleLine(format(LQTr["  |cffffffff%s in|r |cff00ff00%s|r"], mobName, zones), format("|cffffffff%s%%|r", core.questsMobsItems[mobName].items[core.questsObjectives[cValue1][cValue2].desc].dropRate))
					else
						GameTooltip:AddLine(format(LQTr["  |cffffffff%s in|r |cff00ff00%s|r"], mobName, zones))
					end
				end
				GameTooltip:Show()
			elseif core.quests[cValue1] and core.quests[cValue1].id then
				local link = GetQuestLink(core.quests[cValue1].id)
				if link then
					GameTooltip:SetOwner(tracker, anchor)
					GameTooltip:SetHyperlink(link)
				end
			end
		end
	end
end
local questsOnLeave = function(cell)
	GameTooltip:Hide()
end

local cProvider, cPrototype, basecCellPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)
function cPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, disableInteraction, category, cType, cValue1, cValue2, anchor)
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth)
	if not disableInteraction then
		self:EnableMouse(true)
		self.category, self.cType, self.cValue1, self.cValue2, self.anchor = category, cType, cValue1, cValue2, anchor
		self:SetScript("OnEnter", questsOnEnter)
		self:SetScript("OnLeave", questsOnLeave)
		self:SetScript("OnMouseDown", questsOnMouseDown)
	else
		self:EnableMouse(false)
		self.category, self.cType, self.cValue1, self.cValue2, self.anchor = nil, nil, nil, nil, nil
		self:SetScript("OnEnter", nil)
		self:SetScript("OnLeave", nil)
		self:SetScript("OnMouseDown", nil)
	end
	return width, height
end

function cPrototype:ReleaseCell()
	self:EnableMouse(false)
	self.category, self.cType, self.cValue1, self.cValue2, self.anchor = nil, nil, nil, nil, nil
	self:SetScript("OnEnter", nil)
	self:SetScript("OnLeave", nil)
	self:SetScript("OnMouseDown", nil)
end

local function addQuestToTracker(uid, numColumns, disableInteraction, questAdded, replaceHeader, displayTime)
	if chdb.questsTracked[uid] and core.quests[uid] and ((prdb.hideCompletedQuest == 3) or (prdb.hideCompletedQuest == 2 and (core.quests[uid].partyComplete == 0 or core.quests[uid].partyCompleteTime and core.quests[uid].partyCompleteTime + prdb.delayToHideCompletedQuest > displayTime)) or (prdb.hideCompletedQuest == 1 and (core.quests[uid].playerComplete == 0 or core.quests[uid].playerCompleteTime and core.quests[uid].playerCompleteTime + tonumber(prdb.delayToHideCompletedQuest) > displayTime))) and ((prdb.limitQuestLevel and (core.quests[uid].weightedLevel - core.playerLevel <= core.db.profile.quests.maxLevelDiff)) or not prdb.limitQuestLevel) then
		local tempLine = tracker:AddLine()
		local questIcon = prdb.showQuestItemIcon and (core.quests[uid].specialItemIcon and format(" |T%s:%d|t", core.quests[uid].specialItemIcon, prdb.questItemIconSize) or "") or ""
		local numPlayerTag = core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""
		tracker:SetCell(tempLine, 1, format("|cff%s%s|r |cff%s%s|r%s %s", prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor, core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor, core.quests[uid].title, questIcon, numPlayerTag), qTrFontTQ2, "LEFT", numColumns - 1, cProvider, prdb.questsCellPadding, 0, prdb.maxWidth, prdb.maxWidth, disableInteraction, "quests", "quest", uid, nil, trackerAnchor)
		tracker:SetCell(tempLine, numColumns, core.quests[uid].pqStatus, qTrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, prdb.statusMaxWidth, prdb.statusMaxWidth, disableInteraction, "quests", "quest", uid, nil, trackerAnchor)
		if core.questsObjectives[uid] then
			for objective, qoData in pairs(core.questsObjectives[uid]) do
				if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and qoData.partyComplete == 0) or (prdb.hideCompletedObjective == 1 and qoData.playerComplete == 0) then
					tempLine = tracker:AddLine()
					tracker:SetCell(tempLine, 1, format("|cff%s%s|r", prdb.colorizeQuestObjective and qoData.qoColor or prdb.defaultQuestObjectiveColor, qoData.desc), qoTrFontTQ2, "LEFT", numColumns - 1, cProvider, prdb.objectivesCellPadding, 0, prdb.maxWidth, prdb.maxWidth, disableInteraction, "quests", "quest", uid, objective, trackerAnchor)
					tracker:SetCell(tempLine, numColumns, prdb.colorizeQuestObjectiveCounter and qoData.qoStatus or qoData.qoStatusNC, qoTrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, prdb.statusMaxWidth, prdb.statusMaxWidth, disableInteraction, "quests", "quest", uid, objective, trackerAnchor)
				end
			end
		else
			tempLine = tracker:AddLine()
			tracker:SetCell(tempLine, 1, core.quests[uid].objective, qoTrFontTQ2, "LEFT", numColumns - 1, cProvider, prdb.objectivesCellPadding, 0, prdb.maxWidth, prdb.maxWidth, disableInteraction, "quests", "quest", uid, nil, trackerAnchor)
		end
		questAdded = true
		replaceHeader = nil
	end
	return questAdded, replaceHeader
end

local function updateQuests(numColumns, disableInteraction)
	local displayTime = GetTime()
	local tempLine, replaceHeader, questAdded
	if prdb.displayHeader then
		for _, zone in pairs(core.sortedZones) do
			if ((prdb.onlyCurrentZone and zone == currentZone) or not prdb.onlyCurrentZone) and core.zones[zone] then
				chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
				if not replaceHeader then tempLine = tracker:AddLine() end
				replaceHeader = true
				tracker:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r |cff%s%s|r", chdb.zoneDisplay[zone] == 0 and "Plus" or "Minus", prdb.collapseButtonSize, prdb.colorizeHeaderName and core.zones[zone].color or prdb.defaultHeaderNameColor, zone, prdb.colorizeHeaderLevel and core.zones[zone].color or prdb.defaultHeaderLevelColor, prdb.hideZoneLevelRange and "" or core.zones[zone].level), hTrFontTQ2, "LEFT", numColumns -1, cProvider, 0, 0, prdb.maxWidth, prdb.maxWidth, disableInteraction, "quests", "zone", zone, nil, trackerAnchor)
				tracker:SetCell(tempLine, numColumns, " ", hTrFontTQ2, "RIGHT", 1, cProvider, 0, 0, prdb.statusMaxWidth, prdb.statusMaxWidth, disableInteraction, "quests", "zone", zone, nil, trackerAnchor)
				if chdb.zoneDisplay[zone] == 1 then
					for _, uid in pairs(core.sortedQuestsByZone[zone]) do
						questAdded, replaceHeader = addQuestToTracker(uid, numColumns, disableInteraction, questAdded, replaceHeader, displayTime)
					end
				else
					for _, uid in pairs(core.sortedQuestsByZone[zone]) do
						if chdb.questsTracked[uid] and core.quests[uid] and ((prdb.hideCompletedQuest == 3) or (prdb.hideCompletedQuest == 2 and (core.quests[uid].partyComplete == 0 or core.quests[uid].partyCompleteTime and core.quests[uid].partyCompleteTime + prdb.delayToHideCompletedQuest > displayTime)) or (prdb.hideCompletedQuest == 1 and (core.quests[uid].playerComplete == 0 or core.quests[uid].playerCompleteTime and core.quests[uid].playerCompleteTime + tonumber(prdb.delayToHideCompletedQuest) > displayTime))) and ((prdb.limitQuestLevel and (core.quests[uid].weightedLevel - core.playerLevel <= core.db.profile.quests.maxLevelDiff)) or not prdb.limitQuestLevel) then
							replaceHeader = nil
							break
						end
					end
				end
			end
		end
	else
		for _, uid in pairs(core.sortedQuests) do
			if chdb.questsTracked[uid] and ((prdb.onlyCurrentZone and core.quests[uid].zone == currentZone) or not prdb.onlyCurrentZone) and ((prdb.hideCompletedQuest == 3) or (prdb.hideCompletedQuest == 2 and (core.quests[uid].partyComplete == 0 or core.quests[uid].partyCompleteTime and core.quests[uid].partyCompleteTime + prdb.delayToHideCompletedQuest > displayTime)) or (prdb.hideCompletedQuest == 1 and (core.quests[uid].playerComplete == 0 or core.quests[uid].playerCompleteTime and core.quests[uid].playerCompleteTime + tonumber(prdb.delayToHideCompletedQuest) > displayTime))) and ((prdb.limitQuestLevel and (core.quests[uid].weightedLevel - core.playerLevel <= core.db.profile.quests.maxLevelDiff)) or not prdb.limitQuestLevel) then
				questAdded, replaceHeader = addQuestToTracker(uid, numColumns, disableInteraction, questAdded, replaceHeader, displayTime)
			end
		end
	end
	if prdb.showItemsStartingQuest and core.numItemsStartingQuest > 0 then
		local zone = LQTr["Items starting a quest"]
		if not replaceHeader or not tempLine then
			tempLine = tracker:AddLine()
			replaceHeader = nil
		end
		chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
		tracker:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r", chdb.zoneDisplay[zone] == 0 and "Plus" or "Minus", prdb.collapseButtonSize, prdb.colorizeHeaderName and "ffffff" or prdb.defaultHeaderNameColor, zone), hTrFontTQ2, "LEFT", numColumns - 1, cProvider, 0, prdb.rightCellPadding, prdb.maxWidth, prdb.maxWidth, disableInteraction, "quests", "zone", zone, nil, trackerAnchor)
		if chdb.zoneDisplay[zone] == 1 then
			for item, itemData in pairs(core.itemStartingQuest) do
				tempLine = tracker:AddLine()
				tracker:SetCell(tempLine, 1, format("|T%s:%d|t |cff%s%s|r", itemData.icon, prdb.questItemIconSize, itemData.color, itemData.name), qTrFontTQ2, "LEFT", numColumns - 1, cProvider, prdb.questsCellPadding, prdb.rightCellPadding, prdb.maxWidth, prdb.maxWidth, disableInteraction, "quests", "questItem", item, nil, trackerAnchor)
			end
		end
	end
	if prdb.showUsableQuestItemsNotAttached and core.numUsableQuestItems > 0 then
		local zone = LQTr["Usable quest items not linked to a quest"]
		if not replaceHeader or not tempLine then
			tempLine = tracker:AddLine()
			replaceHeader = nil
		end
		chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
		tracker:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r", chdb.zoneDisplay[zone] == 0 and "Plus" or "Minus", prdb.collapseButtonSize, prdb.colorizeHeaderName and "ffffff" or prdb.defaultHeaderNameColor, zone), hTrFontTQ2, "LEFT", numColumns - 1, cProvider, 0, prdb.rightCellPadding, prdb.maxWidth, prdb.maxWidth, disableInteraction, "quests", "zone", zone, nil, trackerAnchor)
		if chdb.zoneDisplay[zone] == 1 then
			for item, itemData in pairs(core.usableQuestItems) do
				tempLine = tracker:AddLine()
				tracker:SetCell(tempLine, 1, format("|T%s:%d|t |cff%s%s|r", itemData.icon, prdb.questItemIconSize, itemData.color, itemData.name), qTrFontTQ2, "LEFT", numColumns - 1, cProvider, prdb.questsCellPadding, prdb.rightCellPadding, prdb.maxWidth, prdb.maxWidth, disableInteraction, "quests", "questItem", item, nil, trackerAnchor)
			end
		end
	end
	if replaceHeader then
		tracker:SetCell(tempLine, 1, nil, hTrFontTQ2, "LEFT", numColumns - 1, cProvider, 0, 0, prdb.maxWidth, prdb.maxWidth, disableInteraction, "inactive")
		tracker:SetCell(tempLine, numColumns, nil, hTrFontTQ2, "RIGHT", 1, cProvider, 0, 0, prdb.statusMaxWidth, prdb.statusMaxWidth, disableInteraction, "inactive")
	end
	if not questAdded and ((prdb.showItemsStartingQuest and core.numItemsStartingQuest == 0) or not prdb.showItemsStartingQuest) then
		return false
	else
		return true
	end
end

function module:updateQuestsTracker(uid)
	if uid and delayedHide[uid] then
		delayedHide[uid] = nil
	end
	local numColumns
	if prdb.toggleTracker and not hideInCombat then
		if core:getModuleEnabled("achievementTracker") and achievementTracker.db.profile.showInQuestTracker then
			local achievementDisplayed, numColumns = achievementTracker:displayAchievement()
			if achievementDisplayed and achievementTracker.db.profile.positionInQuestTracker == 1 then
				tracker = QTIP:Acquire("tomQuest2Tracker", numColumns)
				tracker:Clear()
				tracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
				tracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
				tracker:SetScale(prdb.scale)
				tracker:SetFrameStrata("BACKGROUND")
				achievementTracker:updateAchievement(tracker, trackerAnchor, prdb.disableInteraction, prdb.maxWidth, prdb.statusMaxWidth)
				local tq = updateQuests(numColumns)
			else
				tracker = QTIP:Acquire("tomQuest2Tracker", numColumns)
				tracker:Clear()
				tracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
				tracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
				tracker:SetScale(prdb.scale)
				tracker:SetFrameStrata("BACKGROUND")
				local tq = updateQuests(numColumns)
				if achievementDisplayed then
					achievementTracker:updateAchievement(tracker, trackerAnchor, prdb.disableInteraction, prdb.maxWidth, prdb.statusMaxWidth)
				elseif not tq then
					QTIP:Release(tracker)
					return
				end
			end
		else
			tracker = QTIP:Acquire("tomQuest2Tracker", 2)
			tracker:Clear()
			tracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
			tracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
			tracker:SetScale(prdb.scale)
			tracker:SetFrameStrata("BACKGROUND")
			local tq = updateQuests(2, prdb.disableInteraction)
			if not tq then
				QTIP:Release(tracker)
				return
			end
		end
	else
		QTIP:Release(tracker)
		return
	end
	tracker:ClearAllPoints()
	if prdb.showAnchor then
		tracker:SetPoint(anchorSetPoint[prdb.anchorPosition].point, anchor, anchorSetPoint[prdb.anchorPosition].relPoint)
	else
		tracker:SetPoint(anchorSetPoint[prdb.anchorPosition].point, anchor, anchorSetPoint[prdb.anchorPosition].relPoint, anchorSetPoint[prdb.anchorPosition].x, anchorSetPoint[prdb.anchorPosition].y)
	end
	if not prdb.disableInteraction then
		tracker:UpdateScrolling(prdb.hasMaxHeight and prdb.maxHeight or nil)
		tracker:EnableMouse(true)
	else
		tracker:EnableMouse(false)
	end
	tracker:Show()
end

--------------------------
--- Keybinding globals ---
--------------------------
BINDING_NAME_TQ2QUESTTRACKERTOGGLE = LQTr["Quests Tracker"]

function tQ2toggleQuestTracker()
	prdb.toggleTracker = not prdb.toggleTracker
	if not prdb.toggleTracker then
		QTIP:Release(tracker)
		tracker = nil
	else
		module:updateQuestsTracker()
	end
end