-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("syncParty", "AceEvent-3.0", "AceTimer-3.0", "AceComm-3.0", "AceSerializer-3.0")

-----------------------
--- Library loading ---
-----------------------
local LSP = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-SyncParty")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs, type = pairs, type
local format = string.format
local GetTime = GetTime
local UnitName = UnitName
local GetNumPartyMembers =  GetNumPartyMembers

-----------------
--- variables ---
-----------------
module.modName = LSP["Party Synchronization"]
local addonPrefix = "tomQ2"
local version = 142
local firstCompatibleVersion = 142
local lastIncompatibleVersion = 141
local syncParty = false
local questDetailTimers = {}
local pId = {true, true, true, true}
for i=1, 4 do
	pId[i] = format("party%d", i)
end

function module:OnInitialize()
	self:SetEnabledState(core:getModuleEnabled("syncParty"))
end

function module:OnEnable()
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterComm(addonPrefix, "commHandler")
 	core.RegisterCallback(self, "addPlayerQuest", "sendMessage")
	core.RegisterCallback(self, "updateObjectiveDesc", "sendMessage")
 	core.RegisterCallback(self, "addPlayerQuestObjective", "sendMessage")
	core.RegisterCallback(self, "endPlayerQuest", "sendMessage")
	core.RegisterCallback(self, "playerQuestObjectiveProgress", "sendMessage")
	core.RegisterCallback(self, "playerQuestComplete", "sendMessage")
	core.RegisterCallback(self, "playerUnQuestComplete", "sendMessage")
	core.RegisterCallback(self, "playerQuestFailed", "sendMessage")
	core.RegisterCallback(self, "playerQuestLost", "sendMessage")
	core.RegisterCallback(self, "endSync", "sendMessage")
	core.RegisterCallback(self, "initParty", "PARTY_MEMBERS_CHANGED")
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
end

function module:PARTY_MEMBERS_CHANGED()
	local oldNumPartyMembers = core.numPartyMembers or 0
	core.numPartyMembers = GetNumPartyMembers() or 0
	if core.numPartyMambers == oldNumPartyMembers then return end
	core:computeQuestLevel()
	local timeStamp = GetTime()
	local partySync = 0
	if core.numPartyMembers > 0 then
		for i=1, core.numPartyMembers do
			local partyName = UnitName(pId[i])
			if not core.party[partyName] then
				core.party[partyName] = {}
				core.party[partyName].status = 0
			end
			if not core.party[partyName].timeStamp then
				module:SendCommMessage(addonPrefix, module:Serialize("v", version, firstCompatibleVersion, core.playerClass), "WHISPER", partyName, "BULK")
			end
			core.party[partyName].timeStamp = timeStamp
			partySync = partySync + core.party[partyName].status
		end
	end
	for partyName, partyData in pairs(core.party) do
		if partyData.timeStamp ~= timeStamp then
			for uid, questData in pairs(core.quests) do
				if questData.complete[partyName] then
					core:questLost(partyName, false, uid)
				end
			end
			core.party[partyName] = nil
		end
	end
	if partySync == 0 then
		syncParty = false
	else
		syncParty = true
	end
end

-------------
--- INPUT ---
-------------
local comm = {
	v = function(sender, senderVersion, senderFirstCompatibleVersion, senderClass)
		if senderVersion <= lastIncompatibleVersion then
			core:Print(format(LSP["The tomQuest2 release (%d) used by %s is too old to synchronise with your release (%d) (min %d)"], senderVersion, sender, version, firstCompatibleVersion))
			return
		elseif version < senderFirstCompatibleVersion then
			core:Print(format(LSP["Your tomQuest2 release (%d) is too old to synchorise with release (%d) of %s (min %d)"], version, senderVersion, sender, firstCompatibleVersion))
			return
		elseif version < senderVersion then
			core:Print(format(LSP["New release available (%d) ask %s where to find it"], senderVersion, sender))
		end
		if not core.party[sender] then
			core.party[sender] = {}
		end
		core.party[sender].status = 1
		core.party[sender].class = senderClass
		core.party[sender].color = core.classColors[senderClass]
		core.party[sender].transfer = true
		syncParty = true
		if core.numQuests > 0 then
			module:SendCommMessage(addonPrefix, module:Serialize("w", version, firstCompatibleVersion, core.playerClass), "WHISPER", sender, "BULK")
			for uid, qData in pairs(core.quests) do
				if qData.complete[core.playerName] then
					module:SendCommMessage(addonPrefix, module:Serialize("q", uid, qData.title, qData.level, qData.tag, qData.group, qData.daily, qData.numObjectives, qData.complete[core.playerName], qData.zone), "WHISPER", sender, "BULK")
					local begun = 0
					if core.questsObjectives[uid] then
						for objective, qoData in pairs(core.questsObjectives[uid]) do
							if qoData.got[core.playerName] then
								local got, nGot = qoData.got[core.playerName]:match("^(.-)~(%d-)$")
								module:SendCommMessage(addonPrefix, module:Serialize("o", uid, objective, qoData.desc, qoData.qType, got, qoData.need, nGot, qoData.nNeed), "WHISPER", sender, "BULK")
								if nGot ~= 0 then
									begun = 1
								end
							end
						end
					end
					module:SendCommMessage(addonPrefix, module:Serialize("r", uid, begun), "WHISPER", sender, "BULK")
				end
			end
			module:SendCommMessage(addonPrefix, module:Serialize("d"), "WHISPER", sender, "BULK")
		end
	end,
	w = function(sender, senderVersion, senderFirstCompatibleVersion, senderClass)
		if senderVersion <= lastIncompatibleVersion then
			core:Print(format(LSP["The tomQuest2 release (%d) used by %s is too old to synchronise with your release (%d) (min %d)"], senderVersion, sender, version, firstCompatibleVersion))
			return
		elseif version < senderFirstCompatibleVersion then
			core:Print(format(LSP["Your tomQuest2 release (%d) is too old to synchorise with release (%d) of %s (min %d)"], version, senderVersion, sender, firstCompatibleVersion))
			return
		elseif version < senderVersion then
			core:Print(format(LSP["New release available (%d) ask %s where to find it"], senderVersion, sender))
		end
		if not core.party[sender] then
			core.party[sender] = {}
		end
		core.party[sender].status = 1
		core.party[sender].class = senderClass
		core.party[sender].color = core.classColors[senderClass]
		core.party[sender].transfer = true
		syncParty = true
	end,
	d = function(sender)
		if core.party[sender].status == 1 then
			core.party[sender].transfer = nil
			core.cbh:Fire("questLogUpdated")
		end
	end,
	q = function(sender, uid, title, level, tag, group, daily, numObjectives, completed, zone)
		if not core.party[sender] or core.party[sender].status == 0 then return end
		if not core.quests[uid] and not core.zones[core.partyZone] then
			core:addZone(core.partyZone)
		end
		core:addQuest(sender, false, uid, title, level, tag, group, daily, numObjectives, completed, zone, nil)
		if not core.party[sender].transfer and not (core.quests[uid].extraSync or core.quests[uid].extras) and (not core.quests[uid].objective or not core.quests[uid].description) then
			questDetailTimers[uid] = module:ScheduleTimer("checkQuestObjective", 60, module:Serialize(uid, sender))
		elseif core.party[sender].transfer and not (core.quests[uid].extraSync or core.quests[uid].extras) and (not core.quests[uid].objective or not core.quests[uid].description) then
			core.quests[uid].extraSync = true
			module:SendCommMessage(addonPrefix, module:Serialize("e", uid), "WHISPER", sender, "BULK")
		end
	end,
	o = function(sender, uid, objective, desc, qType, got, need, nGot, nNeed)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.questsObjectives[uid] then return end
		core:addQuestObjective(sender, false, uid, objective, desc, qType, got, need, nGot, nNeed)
	end,
	p = function(sender, uid, objective, got, nGot)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.questsObjectives[uid] or not core.questsObjectives[uid][objective] or not core.questsObjectives[uid][objective].got then return end
		core:questObjectiveProgress(sender, false, uid, objective, got, nGot)
	end,
	r = function(sender, uid, numObjectives, started)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.quests[uid].complete[sender] then return end
		core.cbh:Fire("endPartyQuest", sender, uid, started, core.party[sender].transfer)
	end,
	e = function(sender, uid)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.quests[uid].complete[core.playerName] then return end
		if core.quests[uid].objective then
			module:SendCommMessage(addonPrefix, module:Serialize("n", uid, core.quests[uid].objective), "WHISPER", sender, "BULK")
		end
		if core.quests[uid].description then
			module:SendCommMessage(addonPrefix, module:Serialize("t", uid, core.quests[uid].description), "WHISPER", sender, "BULK")
		end
		module:SendCommMessage(addonPrefix, module:Serialize("g", uid, core.quests[uid].money or 0, core.quests[uid].honor or 0, core.quests[uid].talent or 0, core.quests[uid].titleReward or "", core.quests[uid].numRewards or 0, core.quests[uid].numChoices or 0, core.quests[uid].rewardSpellName or "", core.quests[uid].rewardSpellTexture or ""), "WHISPER", sender, "BULK")
		if core.quests[uid].numRewards and core.quests[uid].numRewards > 0 then
			for i, irData in pairs (core.quests[uid].itemRewards) do
				module:SendCommMessage(addonPrefix, module:Serialize("i", uid, i, irData.name, irData.texture, irData.numItems, irData.quality, irData.isUsable, irData.link, "REWARD"), "WHISPER", sender, "BULK")
			end
		end
		if core.quests[uid].numChoices and core.quests[uid].numChoices > 0 then
			for i, icData in pairs (core.quests[uid].itemChoices) do
				module:SendCommMessage(addonPrefix, module:Serialize("i", uid, i, icData.name, icData.texture, icData.numItems, icData.quality, icData.isUsable, icData.link, "CHOICE"), "WHISPER", sender, "BULK")
			end
		end
	end,
	n = function(sender, uid, objective)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] then return end
		core.quests[uid].objective = objective
		core.cbh:Fire("questObjectiveAdded", uid)
	end,
	t = function(sender, uid, description)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] then return end
		core.quests[uid].description = description
		core.cbh:Fire("questDescriptionAdded", uid)
	end,
	g = function(sender, uid, money, honor, talent, titleReward, numRewards, numChoices, rewardSpellName, rewardSpellTexture)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] then return end
		core.quests[uid].money = money
		core.quests[uid].honor = honor
		core.quests[uid].talent = talent
		core.quests[uid].titleReward = titleReward
		core.quests[uid].numRewards = numRewards
		core.quests[uid].numChoices = numChoices
		core.quests[uid].rewardSpellName = rewardSpellName
		core.quests[uid].rewardSpellTexture = rewardSpellTexture
	end,
	i = function(sender, uid, i, name, texture, numItems, quality, isUsable, link, itype)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] then return end
		if itype == "REWARD" then
			core.quests[uid].itemRewards = core.quests[uid].itemRewards or {}
			if not core.quests[uid].itemRewards[i] then
				core.quests[uid].itemRewards[i] = {}
				core.quests[uid].itemRewards[i].name = name
				core.quests[uid].itemRewards[i].texture = texture
				core.quests[uid].itemRewards[i].numItems = numItems
				core.quests[uid].itemRewards[i].quality = quality
				core.quests[uid].itemRewards[i].isUsable = isUsable
				core.quests[uid].itemRewards[i].link = link
			end
		elseif itype == "CHOICE" then
			core.quests[uid].itemChoices = core.quests[uid].itemChoices or {}
			if not core.quests[uid].itemChoices[i] then
				core.quests[uid].itemChoices[i] = {}
				core.quests[uid].itemChoices[i].name = name
				core.quests[uid].itemChoices[i].texture = texture
				core.quests[uid].itemChoices[i].numItems = numItems
				core.quests[uid].itemChoices[i].quality = quality
				core.quests[uid].itemChoices[i].isUsable = isUsable
				core.quests[uid].itemChoices[i].link = link
			end
		end
	end,
	c = function(sender, uid)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.quests[uid].complete[sender] then return end
		core:questComplete(sender, false, uid)
	end,
	a = function(sender, uid)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.quests[uid].complete[sender] then return end
		core:questUnComplete(sender, false, uid)
	end,
	f = function(sender, uid)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.quests[uid].complete[sender] then return end
		core:questFailed(sender, false, uid)
	end,
	l = function(sender, uid, abandoning)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.quests[uid].complete[sender] then return end
		if questDetailTimers[uid] then
			module:CancelTimer(questDetailTimers[uid], true)
			questDetailTimers[uid] = nil
		end
		core:questLost(sender, false, uid, abandoning, 1)
	end,
}

function module:commHandler(prefix, message, distribution, sender)
	if sender ~= core.playerName then
		module:onCommand(sender, module:Deserialize(message))
	end
end

function module:onCommand(sender, success, commHandler, ...)
	if not success then
		return
	elseif comm[commHandler] then
		comm[commHandler](sender, ...)
	end
end

function module:checkQuestObjective(arg)
	local success, uid, sender = module:Deserialize(arg)
	if not success or not uid or not core.quests[uid] or not sender or not core.party[sender] or core.quests[uid].extraSync or core.quests[uid].extras then return end
	core.quests[uid].extraSync = true
	module:SendCommMessage(addonPrefix, module:Serialize("e", uid), "WHISPER", sender, "BULK")
end
--------------
--- OUTPUT ---
--------------
function module:sendMessage(event, first, commHandler, uid, ...)
	if commHandler == "q" and questDetailTimers[uid] then
		module:CancelTimer(questDetailTimers[uid], true)
		questDetailTimers[uid] = nil
	end
	if core.numPartyMembers > 0 and syncParty and not first then
		module:SendCommMessage(addonPrefix, module:Serialize(commHandler, uid, ...), "PARTY", nil, "BULK")
	end
end